#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import ui
from ... enums import *


UI_LINE_CAP = {
    LINE_CAP_BUTT: ui.LINE_CAP_BUTT,
    LINE_CAP_SQUARE: ui.LINE_CAP_SQUARE,
    LINE_CAP_ROUND: ui.LINE_CAP_ROUND}

UI_LINE_JOIN = {
    LINE_JOIN_BEVEL: ui.LINE_JOIN_BEVEL,
    LINE_JOIN_MITER: ui.LINE_JOIN_MITER,
    LINE_JOIN_ROUND: ui.LINE_JOIN_ROUND}

UI_LINE_STYLE = {
    LINE_STYLE_CUSTOM: LINE_STYLE_CUSTOM,
    LINE_STYLE_SOLID: LINE_STYLE_SOLID,
    LINE_STYLE_DOTTED: DASH_VALUES_DOTTED,
    LINE_STYLE_DASHED: DASH_VALUES_DASHED,
    LINE_STYLE_DASHDOTTED: DASH_VALUES_DASHDOTTED}

UI_FILL_RULE = {
    FILL_RULE_EVENODD: True,
    FILL_RULE_WINDING: False}

UI_FONT_STYLE = {
    FONT_STYLE_NORMAL: FONT_STYLE_NORMAL,
    FONT_STYLE_ITALIC: FONT_STYLE_ITALIC}

UI_FONT_WEIGHT = {
    FONT_WEIGHT_NORMAL: FONT_WEIGHT_NORMAL,
    FONT_WEIGHT_LIGHT: FONT_WEIGHT_LIGHT,
    FONT_WEIGHT_BOLD: FONT_WEIGHT_BOLD,
    FONT_WEIGHT_BLACK: FONT_WEIGHT_BLACK,
    FONT_WEIGHT_HEAVY: FONT_WEIGHT_HEAVY,
    FONT_WEIGHT_SEMIBOLD: FONT_WEIGHT_SEMIBOLD,
    FONT_WEIGHT_MEDIUM: FONT_WEIGHT_MEDIUM,
    FONT_WEIGHT_ULTRALIGHT: FONT_WEIGHT_ULTRALIGHT,
    FONT_WEIGHT_THIN: FONT_WEIGHT_THIN}

UI_FONT_FAMILY = {
    FONT_FAMILY_SERIF: FONT_FACE_SERIF,
    FONT_FAMILY_SANS: FONT_FACE_SANS,
    FONT_FAMILY_MONO: FONT_FACE_MONO}

UI_TOUCH_STATE = {
    'began': TOUCH_PRESSED,
    'ended': TOUCH_RELEASED,
    'moved': TOUCH_MOVED,
    'stationary': TOUCH_STATIC,
    'cancelled': TOUCH_RELEASED}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from . view import ViewEvt


class Touch(object):
    """
    Represents a single touch point.
    
    Attributes:
        
        id: any
            Unique identifier of current touch.
        
        x_pos: int or float
            Logical x-coordinate in device units.
        
        y_pos: int or float
            Logical y-coordinate in device units.
        
        x_prev: int, float or None
            Previous logical x-coordinate in device units.
        
        y_prev: int, float or None
            Previous logical y-coordinate in device units.
        
        force: float
            Touch pressure as a value between 0 and 1.
        
        state: pero.TOUCH
            Specifies current state of the touch as any value from the
            pero.TOUCH enum.
    """
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of Touch."""
        
        self.id = None
        
        self.x_pos = None
        self.y_pos = None
        
        self.x_prev = None
        self.y_prev = None
        
        self.force = None
        self.state = None
        
        # set given arguments
        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)
            else:
                raise AttributeError("Attribute not found! --> %s" % name)


class TouchEvt(ViewEvt):
    """
    Defines a generic event which is fired on any touch-related event.
    
    Attributes:
        
        x_pos: int or float
            Last touch logical x-coordinate in device units.
        
        y_pos: int or float
            Last touch logical y-coordinate in device units.
        
        touches: (pero.Touch,)
            Collection of individual touches.
        
        alt_down: bool
            Indicates Alt key state.
        
        cmd_down: bool
            Indicates Command key state.
        
        ctrl_down: bool
            Indicates Control key state.
        
        shift_down: bool
            Indicates Shift key state.
    """
    
    TYPE = EVT_TOUCH
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of MouseEvt."""
        
        self.touches = None
        
        self.alt_down = None
        self.cmd_down = None
        self.ctrl_down = None
        self.shift_down = None
        
        super().__init__(**kwargs)
    
    
    @property
    def x_pos(self):
        """Gets x-coordinate of the last touch in device units."""
        
        # check touches
        if not self.touches:
            return None
        
        # return last
        return self.touches[-1].x_pos
    
    
    @property
    def y_pos(self):
        """Gets y-coordinate of the last touch in device units."""
        
        # check touches
        if not self.touches:
            return None
        
        # return last
        return self.touches[-1].y_pos
    
    
    @classmethod
    def from_evt(cls, evt):
        """
        Initializes a new instance of given class by copying all data.
        
        Args:
            evt: pero.TouchEvt
                Source event from which to copy the data.
        
        Returns:
            cls instance
                New instance of requested class.
        """
        
        return cls(
            
            native = evt.native,
            view = evt.view,
            control = evt.control,
            
            touches = evt.touches,
            
            alt_down = evt.alt_down,
            cmd_down = evt.cmd_down,
            ctrl_down = evt.ctrl_down,
            shift_down = evt.shift_down)


class TouchStartEvt(TouchEvt):
    """Defines an event which is fired if touch starts."""
    
    TYPE = EVT_TOUCH_START


class TouchEndEvt(TouchEvt):
    """Defines an event which is fired if touch ends."""
    
    TYPE = EVT_TOUCH_END


class TouchMoveEvt(TouchEvt):
    """Defines an event which is fired if touch moves."""
    
    TYPE = EVT_TOUCH_MOVE


class TouchCancelEvt(TouchEvt):
    """Defines an event which is fired if touch cancels."""
    
    TYPE = EVT_TOUCH_CANCEL


class TouchDTapEvt(TouchEvt):
    """Defines an event which is fired on double touch."""
    
    TYPE = EVT_TOUCH_DTAP



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *


class Scale(PropertySet):
    """
    Abstract base class for various types of scales, which are used to convert
    values between two dimensions. Typically they are used in plotting for
    conversion between real data values into device screen units, conversion
    of values into categories or levels etc.
    
    If possible, derived classes should implement conversion methods for both
    directions: the 'scale' method, which converts values from the input range
    to output range and the 'invert' method, which converts from the output
    range to input range.
    
    To convert more data conveniently and/or efficiently, the scales can convert
    not only a single value but whole sequence such as list or numpy.array at
    once. For custom scales be sure this functionality is kept.
    
    Properties:
        
        in_range: (any,), None or UNDEF
            Specifies the input values or range.
        
        out_range: (any,), None or UNDEF
            Specifies the output values or range.
    """
    
    in_range = TupleProperty((), dynamic=False, nullable=True)
    out_range = TupleProperty((), dynamic=False, nullable=True)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return "%s -> %s" % (self.in_range, self.out_range)
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "%s(%s)" % (self.__class__.__name__, self.__str__())
    
    
    def scale(self, value, *args, **kwargs):
        """
        Returns corresponding output-range value for given input-range value.
        
        This method should be overridden in derived classes to provide specific
        mechanism to convert from input to output range.
        """
        
        raise NotImplementedError("The 'scale' method is not implemented for '%s'." % self.__class__.__name__)
    
    
    def invert(self, value, *args, **kwargs):
        """
        Returns corresponding input-range value for given output-range value.
        
        This method should be overridden in derived classes to provide specific
        mechanism to convert from output to input range.
        """
        
        raise NotImplementedError("The 'invert' method is not implemented for '%s'." % self.__class__.__name__)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero
from numpy.testing import assert_array_almost_equal 


class TestCase(unittest.TestCase):
    """Test case for time ticker."""
    
    
    def test_major_ticks_hours(self):
        """Tests whether major ticks are calculated correctly for hours."""
        
        ticker = pero.TimeTicker()
        
        # step 10 hours
        ticker(start=0, end=50*3600, major_count=5)
        ticks = ticker.major_ticks()
        
        model = (0, 10*3600, 20*3600, 30*3600, 40*3600, 50*3600)
        self.assertEqual(ticks, model)
        
        # step 5 hours
        ticker(start=0, end=20*3600, major_count=5)
        ticks = ticker.major_ticks()
        
        model = (0, 5*3600, 10*3600, 15*3600, 20*3600)
        self.assertEqual(ticks, model)
        
        # step 2 hours
        ticker(start=0, end=10*3600, major_count=5)
        ticks = ticker.major_ticks()
        
        model = (0, 2*3600, 4*3600, 6*3600, 8*3600, 10*3600)
        self.assertEqual(ticks, model)
        
        # step 1 hour
        ticker(start=0, end=5*3600, major_count=5)
        ticks = ticker.major_ticks()
        
        model = (0, 1*3600, 2*3600, 3*3600, 4*3600, 5*3600)
        self.assertEqual(ticks, model)
        
        # step 30 min
        ticker(start=0, end=3*3600, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 30*60, 60*60, 90*60, 120*60, 150*60, 180*60)
        self.assertEqual(ticks, model)
        
        # step 15 min
        ticker(start=0, end=2*3600, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 15*60, 30*60, 45*60, 60*60, 75*60, 90*60, 105*60, 120*60)
        self.assertEqual(ticks, model)
        
        # step 10 min
        ticker(start=0, end=3600, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 10*60, 20*60, 30*60, 40*60, 50*60, 60*60)
        self.assertEqual(ticks, model)
    
    
    def test_major_ticks_minutes(self):
        """Tests whether major ticks are calculated correctly for minutes."""
        
        ticker = pero.TimeTicker()
        
        # step 15 min
        ticker(start=0, end=59*60, major_count=4)
        ticks = ticker.major_ticks()
        
        model = (0, 15*60, 30*60, 45*60)
        self.assertEqual(ticks, model)
        
        # step 10 min
        ticker(start=0, end=50*60, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 10*60, 20*60, 30*60, 40*60, 50*60)
        self.assertEqual(ticks, model)
        
        # step 5 min
        ticker(start=0, end=35*60, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 5*60, 10*60, 15*60, 20*60, 25*60, 30*60, 35*60)
        self.assertEqual(ticks, model)
        
        # step 2 min
        ticker(start=0, end=9*60, major_count=6)
        ticks = ticker.major_ticks()
        
        model = (0, 2*60, 4*60, 6*60, 8*60)
        self.assertEqual(ticks, model)
        
        # step 1 min
        ticker(start=0, end=7*60, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 1*60, 2*60, 3*60, 4*60, 5*60, 6*60, 7*60)
        self.assertEqual(ticks, model)
    
    
    def test_major_ticks_seconds(self):
        """Tests whether major ticks are calculated correctly for seconds."""
        
        ticker = pero.TimeTicker()
        
        # step 15 s
        ticker(start=0, end=59, major_count=4)
        ticks = ticker.major_ticks()
        
        model = (0, 15, 30, 45)
        self.assertEqual(ticks, model)
        
        # step 10 s
        ticker(start=0, end=50, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 10, 20, 30, 40, 50)
        self.assertEqual(ticks, model)
        
        # step 5 s
        ticker(start=0, end=35, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 5, 10, 15, 20, 25, 30, 35)
        self.assertEqual(ticks, model)
        
        # step 2 s
        ticker(start=0, end=9, major_count=6)
        ticks = ticker.major_ticks()
        
        model = (0, 2, 4, 6, 8)
        self.assertEqual(ticks, model)
        
        # step 1 s
        ticker(start=0, end=7, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 1, 2, 3, 4, 5, 6, 7)
        self.assertEqual(ticks, model)
    
    
    def test_major_ticks_milliseconds(self):
        """Tests whether major ticks are calculated correctly for milliseconds."""
        
        ticker = pero.TimeTicker()
        
        # step 0.5 s
        ticker(start=0, end=2, major_count=5)
        ticks = ticker.major_ticks()
        
        model = (0, 0.5, 1.0, 1.5, 2.0)
        self.assertEqual(ticks, model)
        
        # step 0.1 s
        ticker(start=0, end=0.9, major_count=7)
        ticks = ticker.major_ticks()
        
        model = (0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9)
        assert_array_almost_equal(ticks, model)
    
    
    def test_beautify(self):
        """Tests whether beautify function works correctly."""
        
        ticker = pero.TimeTicker()
        
        # test hours
        beautified = ticker.beautify(2.1*3600, 5.7*3600)
        self.assertEqual(beautified, (2*3600, 6*3600))
        
        beautified = ticker.beautify(5.7*3600, 2.1*3600)
        self.assertEqual(beautified, (6*3600, 2*3600))
        
        # test minutes
        beautified = ticker.beautify(7*60, 48*60)
        self.assertEqual(beautified, (5*60, 50*60))
        
        beautified = ticker.beautify(48*60, 7*60)
        self.assertEqual(beautified, (50*60, 5*60))


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import cairo
import numpy
from ... properties import *
from ... drawing import Canvas, Path, Matrix, ClipState
from . enums import *


class CairoCanvas(Canvas):
    """Wrapper for Cairo drawing context."""
    
    
    def __init__(self, dc, **overrides):
        """
        Initializes a new instance of CairoCanvas.
        
        Args:
            dc: cairo.Context
                Cairo drawing context.
            
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        # init buffers
        self._dc = dc
        self._pen = (0, 0, 0, 1)
        self._brush = None
        
        self._default_font = self._dc.get_font_face()
        self._font = {
            'family': self._default_font.get_family(),
            'style': self._default_font.get_slant(),
            'weight': self._default_font.get_weight(),
            'for_color': (0, 0, 0, 1),
            'bgr_color': None}
        
        # init base
        super().__init__(**overrides)
        
        # init canvas
        self._update_pen()
        self._update_brush()
        self._update_text()
        
        # bind events
        self.bind(EVT_PEN_CHANGED, self._update_pen)
        self.bind(EVT_BRUSH_CHANGED, self._update_brush)
        self.bind(EVT_TEXT_CHANGED, self._update_text)
    
    
    def get_line_size(self, text):
        """
        Gets width and height of a single text line using current text settings.
        
        Args:
            text: str
                Text for which the size should be calculated.
        
        Returns:
            (float, float)
                Line width and height.
        """
        
        # check text
        if not text:
            return 0, 0
        
        # get size
        width = self._dc.text_extents(text).width
        height = self._dc.font_extents()[2]
        
        return width, height
    
    
    def draw_arc(self, x, y, radius, start_angle, end_angle, clockwise=True):
        """
        Draws an arc of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the arc.
            
            start_angle: int or float
                Angle of the starting point in radians.
            
            end_angle: int or float
                Angle of the starting point in radians.
            
            clockwise: bool
                Direction of drawing between start and end point.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        radius = self._scale * radius
        
        # init drawing
        self._dc.new_path()
        
        # draw
        if clockwise:
            self._dc.arc(x, y, radius, start_angle, end_angle)
        else:
            self._dc.arc_negative(x, y, radius, start_angle, end_angle)
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_circle(self, x, y, radius):
        """
        Draws a circle of specified radius centered around given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        radius = self._scale * radius
        
        # init drawing
        self._dc.new_path()
        
        # draw
        self._dc.arc(x, y, radius, 0, 2*numpy.pi)
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_ellipse(self, x, y, width, height):
        """
        Adds an ellipse centered at given position and fitting into the size.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
        """
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # check size
        if not width or not height:
            return
        
        # get scale and radius
        scale = float(width) / height
        radius = 0.5*height
        
        # init drawing
        self._dc.new_path()
        
        # draw
        self._dc.save()
        self._dc.scale(scale, 1)
        self._dc.arc(x/scale, y, radius, 0, 2*numpy.pi)
        self._dc.restore()
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_line(self, x1, y1, x2, y2):
        """
        Draws a line between two points.
        
        Args:
            x1: int or float
                X-coordinate of the line start.
            
            y1: int or float
                Y-coordinate of the line start.
            
            x2: int or float
                X-coordinate of the line end.
            
            y2: int or float
                Y-coordinate of the line end.
        """
        
        # apply scaling and offset
        x1 = self._scale * (x1 + self._offset[0])
        y1 = self._scale * (y1 + self._offset[1])
        x2 = self._scale * (x2 + self._offset[0])
        y2 = self._scale * (y2 + self._offset[1])
        
        # init drawing
        self._dc.new_path()
        self._dc.move_to(x1, y1)
        
        # draw
        self._dc.line_to(x2, y2)
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_lines(self, points):
        """
        Draws continuous open line using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # init drawing
        self._dc.new_path()
        self._dc.move_to(points[0][0], points[0][1])
        
        # draw
        for p in points[1:]:
            self._dc.line_to(p[0], p[1])
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_path(self, path):
        """
        Draws given path using current pen and brush.
        
        Args:
            path: pero.Path
                Path to be drawn.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # set new fill rule
        fill_rule = self._dc.get_fill_rule()
        self._dc.set_fill_rule(CAIRO_FILL_RULE[path.fill_rule])
        
        # add path
        self._make_native_path(path)
        
        # fill and stroke
        self._fill_and_stroke()
        
        # set back fill rule
        self._dc.set_fill_rule(fill_rule)
    
    
    def draw_polygon(self, points):
        """
        Draws a closed polygon using sequence of points.
        
        Args:
            points: ((float, float),)
                Sequence of x,y coordinates of the points.
        """
        
        # check points
        if len(points) < 2:
            return
        
        # apply scaling and offset
        points = (numpy.array(points) + self._offset) * numpy.array((self._scale, self._scale))
        
        # init drawing
        self._dc.new_path()
        self._dc.move_to(points[0][0], points[0][1])
        
        # draw
        for p in points[1:]:
            self._dc.line_to(p[0], p[1])
        
        self._dc.close_path()
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_rect(self, x, y, width, height, radius=None):
        """
        Draws a rectangle specified by given top left corner and size and
        optional round corners specified as a single value or individual value
        for each corners starting from top-left.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width of the rectangle.
            
            height: int or float
                Full height of the rectangle.
            
            radius: int, float, (int,int,int,int) or (float,float,float,float)
                Radius of curved corners.
        """
        
        # get radius
        if isinstance(radius, (int, float)) and radius != 0:
            radius = (radius, radius, radius, radius)
        
        # draw as path if having round corners
        if radius:
            path = Path()
            path.rect(x, y, width, height, radius)
            self.draw_path(path)
            return
        
        # apply scaling and offset
        x = self._scale * (x + self._offset[0])
        y = self._scale * (y + self._offset[1])
        width = self._scale * width
        height = self._scale * height
        
        # init drawing
        self._dc.new_path()
        
        # draw
        self._dc.rectangle(x, y, width, height)
        
        # fill and stroke
        self._fill_and_stroke()
    
    
    def draw_text(self, text, x, y, angle=0):
        """
        Draws a text string anchored at specified point using current text
        settings.
        
        Args:
            text: str
                Text to be drawn.
            
            x: int or float
                X-coordinate of the text anchor.
            
            y: int or float
                Y-coordinate of the text anchor.
            
            angle: int or float
                Text angle in radians.
        """
        
        # set color
        self._dc.set_source_rgba(*self._font['for_color'])
        
        # get full size
        full_width, full_height = self.get_text_size(text)
        
        # get font size
        ascent, descent, font_height, max_x_advance, max_y_advance = self._dc.font_extents()
        font_height /= self._scale
        
        # apply angle transformation
        if angle:
            
            x = self._scale * (x + self._offset[0])
            y = self._scale * (y + self._offset[1])
            
            self._dc.save()
            self._dc.translate(x, y)
            self._dc.rotate(angle)
            
            x = 0
            y = 0
        
        # split lines
        lines = [text]
        if self.text_split and self.text_splitter:
            lines = text.split(self.text_splitter)
        
        # draw lines
        for i, line in enumerate(lines):
            
            # get line size
            x_bearing, y_bearing, line_width, line_height, x_advance, y_advance = self._dc.text_extents(line)
            line_width /= self._scale
            line_height /= self._scale
            
            # init offset
            x_offset = -x_bearing / self._scale
            y_offset = ascent / self._scale
            
            # adjust alignment
            if self.text_align == TEXT_ALIGN_CENTER:
                x_offset -= 0.5*line_width
            
            elif self.text_align == TEXT_ALIGN_RIGHT:
                x_offset -= line_width
            
            # adjust baseline
            if self.text_base == TEXT_BASE_MIDDLE:
                y_offset -= 0.5*full_height
            
            elif self.text_base == TEXT_BASE_BOTTOM:
                y_offset -= full_height
            
            # add line offset
            y_offset += i * font_height * (1 + self.text_spacing)
            
            # apply scaling and offset
            if angle:
                line_x = self._scale * x_offset
                line_y = self._scale * y_offset
            else:
                line_x = self._scale * (x + x_offset + self._offset[0])
                line_y = self._scale * (y + y_offset + self._offset[1])
            
            # draw background
            if self._font['bgr_color']:
                
                bgr_x = line_x + x_bearing
                bgr_y = line_y - ascent
                bgr_width = line_width * self._scale
                bgr_height = font_height * self._scale
                
                self._dc.set_source_rgba(*self._font['bgr_color'])
                self._dc.rectangle(bgr_x, bgr_y, bgr_width, bgr_height)
                self._dc.fill()
                self._dc.set_source_rgba(*self._font['for_color'])
            
            # draw text
            self._dc.move_to(line_x, line_y)
            self._dc.show_text(line)
        
        # revert angle transformation
        if angle:
            self._dc.restore()
    
    
    def clip(self, path):
        """
        Sets clipping path as intersection with current one.
        
        Args:
            path: pero.Path
                Path to be used for clipping.
        
        Returns:
            pero.ClipState
                Clipping state context manager.
        """
        
        # apply scaling and offset
        matrix = Matrix()
        matrix.translate(self._offset[0], self._offset[1])
        matrix.scale(self._scale, self._scale)
        path = path.transformed(matrix)
        
        # save current canvas state
        self._dc.save()
        
        # add path
        self._make_native_path(path)
        
        # set as clipping
        self._dc.clip()
        
        # return state
        return ClipState(self)
    
    
    def unclip(self):
        """Removes last clipping path while keeping previous if any."""
        
        # restore canvas state
        self._dc.restore()
        
        # set pen, brush and text
        self._update_pen()
        self._update_brush()
        self._update_text()
    
    
    def _make_native_path(self, path):
        """Adds given path to canvas without drawing it."""
        
        # init path
        self._dc.new_path()
        
        # draw
        for command in path.commands():
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                self._dc.close_path()
            
            # move to
            elif key == PATH_MOVE:
                self._dc.move_to(*values)
            
            # line to
            elif key == PATH_LINE:
                self._dc.line_to(*values)
            
            # curve to
            elif key == PATH_CURVE:
                self._dc.curve_to(*values)
    
    
    def _fill_and_stroke(self):
        """Fills and strokes current path."""
        
        # fill and stroke
        if self._pen and self._brush:
            
            self._dc.set_source_rgba(*self._brush)
            self._dc.fill_preserve()
            
            self._dc.set_source_rgba(*self._pen)
            self._dc.stroke()
        
        # stroke only
        elif self._pen:
            
            self._dc.set_source_rgba(*self._pen)
            self._dc.stroke()
        
        # fill only
        elif self._brush:
            
            self._dc.set_source_rgba(*self._brush)
            self._dc.fill()
    
    
    def _update_pen(self, evt=None):
        """Updates pen with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # update color
        if prop_name is None or prop_name in ('line_color', 'line_alpha'):
            color = ColorProperties.get_color(self, "line_")
            if color is not UNDEF:
                self._pen = color.rgba_r if color.alpha else None
        
        # update width
        if prop_name is None or prop_name in ('line_width', 'line_scale'):
            line_width = self.line_width
            if line_width is not UNDEF:
                self._dc.set_line_width(line_width * self.line_scale)
        
        # update cap
        if prop_name is None or prop_name == 'line_cap':
            line_cap = self.line_cap
            if line_cap is not UNDEF:
                self._dc.set_line_cap(CAIRO_LINE_CAP[line_cap])
        
        # update join
        if prop_name is None or prop_name == 'line_join':
            line_join = self.line_join
            if line_join is not UNDEF:
                self._dc.set_line_join(CAIRO_LINE_JOIN[line_join])
        
        # update style/dash
        if prop_name is None or prop_name in ('line_dash', 'line_style', 'line_width', 'line_scale'):
            line_style = self.line_style
            line_dash = self.line_dash if self.line_dash else []
            line_width = self._dc.get_line_width()
            
            if line_style == LINE_STYLE_SOLID:
                line_dash = []
            elif line_style not in (LINE_STYLE_CUSTOM, UNDEF):
                line_dash = CAIRO_LINE_STYLE[line_style]
            
            self._dc.set_dash([x*line_width for x in line_dash])
    
    
    def _update_brush(self, evt=None):
        """Updates brush with current properties."""
        
        color = ColorProperties.get_color(self, "fill_")
        
        if self.fill_style == FILL_STYLE_TRANS:
            self._brush = None
        
        elif color is not UNDEF:
            self._brush = color.rgba_r if color.alpha else None
    
    
    def _update_text(self, evt=None):
        """Updates text with current properties."""
        
        # get property name
        prop_name = evt.name if evt is not None else None
        
        # init flag
        init_font = False
        
        # update font name/family
        if prop_name is None or prop_name in ('font_name', 'font_family'):
            font_name = self.font_name
            font_family = self.font_family
            
            if font_name:
                self._font['family'] = font_name
            elif font_family is None:
                self._font['family'] = self._default_font.get_family()
            elif font_family is not UNDEF:
                self._font['family'] = CAIRO_FONT_FAMILY[font_family]
            
            init_font = True
        
        # update font size
        if prop_name is None or prop_name in ('font_size', 'font_scale'):
            font_size = self.font_size
            
            if font_size is None:
                self._dc.set_font_size(11 * self.font_scale)
            elif font_size is not UNDEF:
                self._dc.set_font_size(font_size * self.font_scale)
        
        # update font style
        if prop_name is None or prop_name == 'font_style':
            font_style = self.font_style
            
            if font_style is None:
                self._font['style'] = self._default_font.get_slant()
            elif font_style is not UNDEF:
                self._font['style'] = CAIRO_FONT_STYLE[font_style]
            
            init_font = True
        
        # update font weight
        if prop_name is None or prop_name == 'font_weight':
            font_weight = self.font_weight
            
            if font_weight is None:
                self._font['weight'] = self._default_font.get_weight()
            elif font_weight is not UNDEF:
                self._font['weight'] = CAIRO_FONT_WEIGHT[font_weight]
            
            init_font = True
        
        # update foreground color
        if prop_name is None or prop_name in ('text_color', 'text_alpha'):
            color = ColorProperties.get_color(self, "text_")
            if color is not UNDEF:
                self._font['for_color'] = color.rgba_r
        
        # update background color
        if prop_name is None or prop_name in ('text_bgr_color', 'text_bgr_alpha'):
            color = ColorProperties.get_color(self, "text_bgr_")
            if color is not UNDEF:
                self._font['bgr_color'] = color.rgba_r if color.alpha else None
        
        # set_font
        if init_font:
            self._dc.select_font_face(
                self._font['family'],
                self._font['style'],
                self._font['weight'])



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for Frame class."""
    
    
    def test_init(self):
        """Tests whether constructor works correctly."""
        
        frame = pero.Frame(10, 20, 100, 200)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 200)
        
        self.assertEqual(frame.x1, 10)
        self.assertEqual(frame.y1, 20)
        self.assertEqual(frame.x2, 110)
        self.assertEqual(frame.y2, 220)
        
        self.assertEqual(frame.cx, 60)
        self.assertEqual(frame.cy, 120)
        
        self.assertEqual(frame.tl, (10, 20))
        self.assertEqual(frame.tr, (110, 20))
        self.assertEqual(frame.bl, (10, 220))
        self.assertEqual(frame.br, (110, 220))
        self.assertEqual(frame.c, (60, 120))
        
        self.assertEqual(frame.w, 100)
        self.assertEqual(frame.h, 200)
        self.assertEqual(frame.wh, (100, 200))
        self.assertEqual(frame.rect, (10, 20, 100, 200))
    
    
    def test_offset(self):
        """Tests whether offset works correctly."""
        
        frame = pero.Frame(10, 20, 100, 200)
        frame.offset(x=10, y=10)
        
        self.assertEqual(frame.x, 20)
        self.assertEqual(frame.y, 30)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 200)
    
    
    def test_shrink(self):
        """Tests whether shrinking works correctly."""
        
        frame = pero.Frame(10, 20, 100, 200)
        frame.shrink(1, 2, 3, 4)
        
        self.assertEqual(frame.x1, 14)
        self.assertEqual(frame.y1, 21)
        self.assertEqual(frame.x2, 108)
        self.assertEqual(frame.y2, 217)
        self.assertEqual(frame.width, 94)
        self.assertEqual(frame.height, 196)
    
    
    def test_extend_coordinate(self):
        """Tests whether extension by single coordinate works correctly."""
        
        # test inside x
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=80)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 200)
        
        # test left x
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=5)
        
        self.assertEqual(frame.x, 5)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 105)
        self.assertEqual(frame.height, 200)
        
        # test right x
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=120)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 110)
        self.assertEqual(frame.height, 200)
        
        # test inside y
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(y=120)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 200)
        
        # test top y
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(y=15)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 15)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 205)
        
        # test bottom y
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(y=250)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 230)
    
    
    def test_extend_point(self):
        """Tests whether extension by point works correctly."""
        
        # test inside
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=80, y=120)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 200)
        
        # test top
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=80, y=15)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 15)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 205)
        
        # test right
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=120, y=120)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 110)
        self.assertEqual(frame.height, 200)
        
        # test bottom
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=80, y=250)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 100)
        self.assertEqual(frame.height, 230)
        
        # test left
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=5, y=120)
        
        self.assertEqual(frame.x, 5)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 105)
        self.assertEqual(frame.height, 200)
        
        # test top right
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=120, y=15)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 15)
        self.assertEqual(frame.width, 110)
        self.assertEqual(frame.height, 205)
        
        # test bottom right
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=120, y=250)
        
        self.assertEqual(frame.x, 10)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 110)
        self.assertEqual(frame.height, 230)
        
        # test bottom left
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=5, y=250)
        
        self.assertEqual(frame.x, 5)
        self.assertEqual(frame.y, 20)
        self.assertEqual(frame.width, 105)
        self.assertEqual(frame.height, 230)
        
        # test top left
        frame = pero.Frame(10, 20, 100, 200)
        frame.extend(x=5, y=15)
        
        self.assertEqual(frame.x, 5)
        self.assertEqual(frame.y, 15)
        self.assertEqual(frame.width, 105)
        self.assertEqual(frame.height, 205)
    
    
    def test_union(self):
        """Tests whether union works correctly."""
        
        # init frame
        frame = pero.Frame(100, 100, 100, 100)
        
        # no overlap
        test = pero.Frame(0, 0, 10, 10)
        result = frame.union(test)
        
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, 0)
        self.assertEqual(result.width, 200)
        self.assertEqual(result.height, 200)
        
        test = pero.Frame(200, 150, 100, 100)
        result = frame.union(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 200)
        self.assertEqual(result.height, 150)
        
        test = pero.Frame(150, 200, 100, 100)
        result = frame.union(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 150)
        self.assertEqual(result.height, 200)
        
        # full cover
        test = pero.Frame(0, 0, 200, 200)
        result = frame.union(test)
        
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, 0)
        self.assertEqual(result.width, 200)
        self.assertEqual(result.height, 200)
        
        # full inside
        test = pero.Frame(125, 125, 50, 50)
        result = frame.union(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 100)
        self.assertEqual(result.height, 100)
        
        # horizontal
        test = pero.Frame(0, 125, 300, 50)
        result = frame.union(test)
        
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 300)
        self.assertEqual(result.height, 100)
        
        # vertical
        test = pero.Frame(125, 0, 50, 300)
        result = frame.union(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 0)
        self.assertEqual(result.width, 100)
        self.assertEqual(result.height, 300)
        
        # top-left
        test = pero.Frame(0, 0, 150, 150)
        result = frame.union(test)
        
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, 0)
        self.assertEqual(result.width, 200)
        self.assertEqual(result.height, 200)
        
        # top-right
        test = pero.Frame(150, 150, 100, 100)
        result = frame.union(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 150)
        self.assertEqual(result.height, 150)
        
        # bottom-left
        test = pero.Frame(0, 150, 150, 150)
        result = frame.union(test)
        
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 200)
        self.assertEqual(result.height, 200)
        
        # bottom-right
        test = pero.Frame(150, 150, 100, 100)
        result = frame.union(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 150)
        self.assertEqual(result.height, 150)
    
    
    def test_intersection(self):
        """Tests whether intersection works correctly."""
        
        # init frame
        frame = pero.Frame(100, 100, 100, 100)
        
        # no overlap
        test = pero.Frame(0, 0, 10, 10)
        result = frame.intersection(test)
        self.assertTrue(result is None)
        
        test = pero.Frame(200, 150, 100, 100)
        result = frame.intersection(test)
        self.assertTrue(result is None)
        
        test = pero.Frame(150, 200, 100, 100)
        result = frame.intersection(test)
        self.assertTrue(result is None)
        
        # full cover
        test = pero.Frame(0, 0, 200, 200)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 100)
        self.assertEqual(result.height, 100)
        
        # full inside
        test = pero.Frame(125, 125, 50, 50)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 125)
        self.assertEqual(result.y, 125)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 50)
        
        # horizontal
        test = pero.Frame(0, 125, 300, 50)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 125)
        self.assertEqual(result.width, 100)
        self.assertEqual(result.height, 50)
        
        # vertical
        test = pero.Frame(125, 0, 50, 300)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 125)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 100)
        
        # top-left
        test = pero.Frame(0, 0, 150, 150)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 100)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 50)
        
        # top-right
        test = pero.Frame(150, 150, 100, 100)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 150)
        self.assertEqual(result.y, 150)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 50)
        
        # bottom-left
        test = pero.Frame(0, 150, 150, 150)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 100)
        self.assertEqual(result.y, 150)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 50)
        
        # bottom-right
        test = pero.Frame(150, 150, 100, 100)
        result = frame.intersection(test)
        
        self.assertEqual(result.x, 150)
        self.assertEqual(result.y, 150)
        self.assertEqual(result.width, 50)
        self.assertEqual(result.height, 50)
    
    
    def test_contains(self):
        """Tests whether contains check works correctly."""
        
        # init frame
        frame = pero.Frame(10, 20, 100, 200)
        
        # test inside x
        self.assertTrue(frame.contains(x=50, y=50))
        
        # test left x
        self.assertFalse(frame.contains(x=5, y=50))
        
        # test right x
        self.assertFalse(frame.contains(x=120, y=50))
        
        # test inside y
        self.assertTrue(frame.contains(x=50, y=50))
        
        # test top y
        self.assertFalse(frame.contains(x=50, y=15))
        
        # test bottom y
        self.assertFalse(frame.contains(x=50, y=250))
    
    
    def test_overlaps(self):
        """Tests whether overlaps check works correctly."""
        
        # init frame
        frame = pero.Frame(100, 100, 100, 100)
        
        # test top-left out
        test = pero.Frame(0, 0, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test top-right out
        test = pero.Frame(300, 0, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test bottom-right out
        test = pero.Frame(300, 300, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test bottom-left out
        test = pero.Frame(0, 300, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test top out
        test = pero.Frame(110, 0, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test right out
        test = pero.Frame(300, 110, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test bottom out
        test = pero.Frame(110, 300, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test left out
        test = pero.Frame(0, 110, 10, 10)
        self.assertFalse(frame.overlaps(test))
        self.assertFalse(test.overlaps(frame))
        
        # test inside
        test = pero.Frame(110, 110, 10, 10)
        self.assertTrue(frame.overlaps(test))
        self.assertTrue(test.overlaps(frame))
        
        # test top-left in
        test = pero.Frame(0, 0, 110, 110)
        self.assertTrue(frame.overlaps(test))
        self.assertTrue(test.overlaps(frame))
        
        # test top-right in
        test = pero.Frame(195, 0, 110, 110)
        self.assertTrue(frame.overlaps(test))
        self.assertTrue(test.overlaps(frame))
        
        # test bottom-right in
        test = pero.Frame(195, 195, 110, 110)
        self.assertTrue(frame.overlaps(test))
        self.assertTrue(test.overlaps(frame))
        
        # test bottom-left in
        test = pero.Frame(0, 195, 110, 110)
        self.assertTrue(frame.overlaps(test))
        self.assertTrue(test.overlaps(frame))


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import fitz
from . enums import *
from . canvas import MuPDFCanvas


def export(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as PDF document into specified file.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
    """
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # init document
    doc = fitz.open()
    page = doc.new_page(width=width, height=height)
    
    # init canvas
    canvas = MuPDFCanvas(page)
    
    if 'draw_scale' in options:
        canvas.draw_scale = options['draw_scale']
    
    if 'line_scale' in options:
        canvas.line_scale = options['line_scale']
    
    if 'font_scale' in options:
        canvas.font_scale = options['font_scale']
    
    # draw graphics
    graphics.draw(canvas)
    
    # save to file
    doc.save(path)
    doc.close()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for Linear scale class."""
    
    
    def test_scale(self):
        """Tests whether scale function works correctly."""
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (10, 350))

        self.assertEqual(scale.scale(50), -32.5)
        self.assertEqual(scale.scale(100), 10)
        self.assertEqual(scale.scale(300), 180)
        self.assertEqual(scale.scale(500), 350)
        self.assertEqual(scale.scale(600), 435)
    
    
    def test_invert(self):
        """Tests whether invert function works correctly."""
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (10, 350))

        self.assertEqual(scale.invert(-32.5), 50)
        self.assertEqual(scale.invert(10), 100)
        self.assertEqual(scale.invert(180), 300)
        self.assertEqual(scale.invert(350), 500)
        self.assertEqual(scale.invert(435), 600)
    
    
    def test_clip(self):
        """Tests whether clipping works correctly."""
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (10, 350),
            clip = True)
        
        self.assertEqual(scale.scale(10), 10)
        self.assertEqual(scale.scale(100), 10)
        self.assertEqual(scale.scale(300), 180)
        self.assertEqual(scale.scale(500), 350)
        self.assertEqual(scale.scale(600), 350)

        self.assertEqual(scale.invert(-32.5), 100)
        self.assertEqual(scale.invert(10), 100)
        self.assertEqual(scale.invert(180), 300)
        self.assertEqual(scale.invert(350), 500)
        self.assertEqual(scale.invert(435), 500)
        
        scale = pero.LinScale(
            in_range = (500, 100),
            out_range = (350, 10),
            clip = True)
        
        self.assertEqual(scale.scale(10), 10)
        self.assertEqual(scale.scale(100), 10)
        self.assertEqual(scale.scale(300), 180)
        self.assertEqual(scale.scale(500), 350)
        self.assertEqual(scale.scale(600), 350)

        self.assertEqual(scale.invert(-32.5), 100)
        self.assertEqual(scale.invert(10), 100)
        self.assertEqual(scale.invert(180), 300)
        self.assertEqual(scale.invert(350), 500)
        self.assertEqual(scale.invert(435), 500)
        
        scale = pero.LinScale(
            in_range = (500, 100),
            out_range = (10, 350),
            clip = True)
        
        self.assertEqual(scale.scale(10), 350)
        self.assertEqual(scale.scale(100), 350)
        self.assertEqual(scale.scale(300), 180)
        self.assertEqual(scale.scale(500), 10)
        self.assertEqual(scale.scale(600), 10)

        self.assertEqual(scale.invert(-32.5), 500)
        self.assertEqual(scale.invert(10), 500)
        self.assertEqual(scale.invert(180), 300)
        self.assertEqual(scale.invert(350), 100)
        self.assertEqual(scale.invert(435), 100)
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (350, 10),
            clip = True)
        
        self.assertEqual(scale.scale(10), 350)
        self.assertEqual(scale.scale(100), 350)
        self.assertEqual(scale.scale(300), 180)
        self.assertEqual(scale.scale(500), 10)
        self.assertEqual(scale.scale(600), 10)

        self.assertEqual(scale.invert(-32.5), 500)
        self.assertEqual(scale.invert(10), 500)
        self.assertEqual(scale.invert(180), 300)
        self.assertEqual(scale.invert(350), 100)
        self.assertEqual(scale.invert(435), 100)
    
    
    def test_arrays(self):
        """Tests whether scale works correctly with arrays."""
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (10, 350))
        
        data = [50, 100, 300, 500, 600]
        model = [-32.5, 10, 180, 350, 435]
        
        self.assertEqual(list(scale.scale(data)), model)
        self.assertEqual(list(scale.invert(model)), data)
    
    
    def test_arrays_clip(self):
        """Tests whether scale clipping works correctly with arrays."""
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (10, 350),
            clip = True)
        
        data = [10, 100, 300, 500, 600]
        model = [10, 10, 180, 350, 350]
        self.assertEqual(list(scale.scale(data)), model)
        
        data = [-32.5, 10, 180, 350, 435]
        model = [100, 100, 300, 500, 500]
        self.assertEqual(list(scale.invert(data)), model)
        
        scale = pero.LinScale(
            in_range = (500, 100),
            out_range = (350, 10),
            clip = True)
        
        data = [10, 100, 300, 500, 600]
        model = [10, 10, 180, 350, 350]
        self.assertEqual(list(scale.scale(data)), model)
        
        data = [-32.5, 10, 180, 350, 435]
        model = [100, 100, 300, 500, 500]
        self.assertEqual(list(scale.invert(data)), model)
        
        scale = pero.LinScale(
            in_range = (500, 100),
            out_range = (10, 350),
            clip = True)
        
        data = [10, 100, 300, 500, 600]
        model = [350, 350, 180, 10, 10]
        self.assertEqual(list(scale.scale(data)), model)
        
        data = [-32.5, 10, 180, 350, 435]
        model = [500, 500, 300, 100, 100]
        self.assertEqual(list(scale.invert(data)), model)
        
        scale = pero.LinScale(
            in_range = (100, 500),
            out_range = (350, 10),
            clip = True)
        
        data = [10, 100, 300, 500, 600]
        model = [350, 350, 180, 10, 10]
        self.assertEqual(list(scale.scale(data)), model)
        
        data = [-32.5, 10, 180, 350, 435]
        model = [500, 500, 300, 100, 100]
        self.assertEqual(list(scale.invert(data)), model)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from . view import ViewEvt


class KeyEvt(ViewEvt):
    """
    Defines a generic event which is fired on any key-related event.
    
    Attributes:
        
        key: int
            Unicode key.
        
        char: str or None
            Character string or None if not character.
        
        pressed: bool
            Indicates the key pressed state.
        
        alt_down: bool
            Indicates Alt key state.
        
        cmd_down: bool
            Indicates Command key state.
        
        ctrl_down: bool
            Indicates Control key state.
        
        shift_down: bool
            Indicates Shift key state.
    """
    
    TYPE = EVT_KEY
    
    def __init__(self, **kwargs):
        """Initializes a new instance of KeyEvt."""
        
        self.key = None
        self.char = None
        self.pressed = None
        
        self.alt_down = None
        self.cmd_down = None
        self.ctrl_down = None
        self.shift_down = None
        
        super().__init__(**kwargs)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        # init from base
        text = super().__str__()
        
        # add key
        text += " key: %s" % self.key
        
        # add char
        if self.char:
            text += " char: %s" % self.char
        
        # add modifiers
        if self.alt_down:
            text += " Alt"
        if self.cmd_down:
            text += " Cmd"
        if self.ctrl_down:
            text += " Ctrl"
        if self.shift_down:
            text += " Shift"
        
        return text
    
    
    @classmethod
    def from_evt(cls, evt):
        """
        Initializes a new instance of given class by copying all data.
        
        Args:
            evt: pero.KeyEvt
                Source event from which to copy the data.
        
        Returns:
            cls instance
                New instance of requested class.
        """
        
        return cls(
            
            native = evt.native,
            view = evt.view,
            control = evt.control,
            
            key = evt.key,
            char = evt.char,
            pressed = evt.pressed,
            
            alt_down = evt.alt_down,
            cmd_down = evt.cmd_down,
            ctrl_down = evt.ctrl_down,
            shift_down = evt.shift_down)


class KeyDownEvt(KeyEvt):
    """Defines an event which is fired if a key is pressed."""
    
    TYPE = EVT_KEY_DOWN


class KeyUpEvt(KeyEvt):
    """Defines an event which is fired if a key is released."""
    
    TYPE = EVT_KEY_UP



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. colors import Color, Palette, Gradient
from . undefined import UNDEF
from . prop import Property


class ColorProperty(Property):
    """
    Defines a color property, which simplifies a color definition by
    automatically creating a pero.Color instance from various input options
    such as an RGB(A) channels (0-255 each), hex string (including leading '#'
    sign) or registered name.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of ColorProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Color, str, tuple, list)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check color
        if isinstance(value, Color):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # check func
        if callable(value):
            return value
        
        # convert to color
        return Color.create(value)


class PaletteProperty(Property):
    """
    Defines a color palette property, which simplifies a palette definition by
    automatically creating a pero.Palette instance from various input options
    such as a list of supported pero.Color definitions or registered palette
    name.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of PaletteProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Palette, str, tuple, list)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check palette
        if isinstance(value, Palette):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # check func
        if callable(value):
            return value
        
        # convert to palette
        return Palette.create(value)


class GradientProperty(Property):
    """
    Defines a color gradient property, which simplifies a gradient definition by
    automatically creating a pero.Gradient instance from various input
    options such as a list of supported pero.Color definitions, pero.Palette
    or registered palette or gradient name.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of GradientProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Gradient, Palette, str, tuple, list)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check gradient
        if isinstance(value, Gradient):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # allow functions
        if callable(value):
            return value
        
        # convert to gradient
        return Gradient.create(value)


class DashProperty(Property):
    """
    Defines a line dash property. The value must be provided as a list or tuple
    of numbers defining the length of lines and spaces in-between.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of DashProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (tuple, list)
        
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check type
        if isinstance(value, (list,tuple)) and all(isinstance(x, (int, float)) for x in value):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow None or UNDEF
        if value is None or value is UNDEF:
            return value
        
        # allow functions
        if callable(value):
            return value
        
        # wrong value
        message = "Value of '%s' property must be a list or a tuple of numbers! -> %s" % (self.name, value)
        raise TypeError(message)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from . values import *
from . enum import Enum

# define pi
PI2X = math.pi*2  # 360
PI = math.pi  # 180
PI2 = math.pi/2  # 90
PI3 = math.pi/3  # 60
PI4 = math.pi/4  # 45
PI6 = math.pi/6  # 30
PI12 = math.pi/12  # 15

# define angle units
ANGLE_DEG = DEG
ANGLE_RAD = RAD

ANGLE = Enum(
    DEG = ANGLE_DEG,
    RAD = ANGLE_RAD)

# define rounding
ROUND_CEIL = CEIL
ROUND_FLOOR = FLOOR
ROUND_HALFUP = HALFUP

ROUNDING = Enum(
    CEIL = ROUND_CEIL,
    FLOOR = ROUND_FLOOR,
    HALFUP = ROUND_HALFUP)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from . library import Library

# init library
COLORS = Library()


class ColorMeta(type):
    """Defines a meta class for the main Color class."""
    
    
    def __getattr__(cls, key):
        """
        Gets registered named color by its name.
        
        Returns:
            color: pero.Color
                Corresponding color.
        """
        
        return COLORS[key]


class Color(object, metaclass=ColorMeta):
    """
    Represents a color defined by red, green, blue and alpha channels. All the
    channels are defined as integers between 0 to 255.
    """
    
    
    def __init__(self, *args, name=None):
        """
        Initializes a new instance of Color.
        
        Args:
            args: (int, int, int), (int, int, int, int) or str
                The RGB(A) channels can be provided as 3 or 4 integers for
                individual channels, tuple of 3 or 4 integers or hex string with
                leading '#'.
            
            name: str or None
                Unique name to register.
        """
        
        # init channels
        channels = None
        red, green, blue, alpha = (0, 0, 0, 255)
        
        # convert from channels
        if len(args) == 3 or len(args) == 4:
            channels = args
        
        # convert from single value
        elif len(args) == 1:
            
            # get value
            value = args[0]
            
            # get channels
            if isinstance(value, (list, tuple)):
                channels = value
            
            # convert from hex
            elif isinstance(value, str):
                
                # strip prefix
                value = value.lstrip('#')
                
                # parse channels
                if len(value) == 3 or len(value) == 4:
                    channels = list(int(value[i]+value[i], 16) for i in range(0, len(value)))
                
                elif len(value) == 6 or len(value) == 8:
                    channels = list(int(value[i:i+2], 16) for i in range(0, len(value), 2))
        
        # check channels
        if channels is None or len(channels) < 3 or len(channels) > 4:
            message = "Unrecognized color definition! -> %s" % (args,)
            raise ValueError(message)
        
        # get channels
        if len(channels) == 3:
            red, green, blue = channels
        
        elif len(channels) == 4:
            red, green, blue, alpha = channels
        
        # check channels
        for c in (red, green, blue, alpha):
            if not isinstance(c, (int, float)) or c < 0 or c > 255:
                message = "Color channels must be a number between 0 and 255! -> (%s, %s, %s, %s)" % (red, green, blue, alpha)
                raise ValueError(message)
        
        # set values
        self._red = int(0.5 + red)
        self._green = int(0.5 + green)
        self._blue = int(0.5 + blue)
        self._alpha = int(0.5 + alpha)
        self._name = name
        
        # register color by name
        if name is not None:
            COLORS.add(self)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        name = ""
        if self._name is not None:
            name = " %s" % self._name
        
        return "(%s,%s,%s,%s)%s" % (self._red, self._green, self._blue, self._alpha, name)
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "%s(%s)" % (self.__class__.__name__, self.__str__())
    
    
    def __eq__(self, other):
        """Equal operator."""
        
        if self is other:
            return True
        
        if isinstance(other, Color):
            return (self.red == other.red
                and self.green == other.green
                and self.blue == other.blue
                and self.alpha == other.alpha)
        
        if isinstance(other, (list, tuple)):
            
            if len(other) == 3:
                return (self.red == other[0]
                    and self.green == other[1]
                    and self.blue == other[2]
                    and self.alpha == 255)
            
            if len(other) == 4:
                return (self.red == other[0]
                    and self.green == other[1]
                    and self.blue == other[2]
                    and self.alpha == other[3])
        
        return False
    
    
    def __ne__(self, other):
        """Not equal operator."""
        
        return not self.__eq__(other)
    
    
    @property
    def name(self):
        """
        Gets color name.
        
        Returns:
            str or None
                Color name.
        """
        
        return self._name
    
    
    @property
    def red(self):
        """
        Gets red channel as a value in range 0 to 255.
        
        Returns:
            int
                Red channel value.
        """
        
        return self._red
    
    
    @property
    def green(self):
        """
        Gets green channel as a value in range 0 to 255.
        
        Returns:
            int
                Green channel value.
        """
        
        return self._green
    
    
    @property
    def blue(self):
        """
        Gets blue channel as a value in range 0 to 255.
        
        Returns:
            int
                Blue channel value.
        """
        
        return self._blue
    
    
    @property
    def alpha(self):
        """
        Gets alpha channel as a value in range 0 to 255 where 0 means fully
        transparent and 255 fully opaque.
        
        Returns:
            int
                Alpha channel value.
        """
        
        return self._alpha
    
    
    @property
    def rgba(self):
        """
        Gets RGBA channels tuple where each channel is defined as integer in
        range 0 to 255.
        
        Returns:
            (int, int, int, int)
                Values of RGBA channels.
        """
        
        return self._red, self._green, self._blue, self._alpha
    
    
    @property
    def rgb(self):
        """
        Gets RGB channels tuple where each channel is defined as integer in
        range 0 to 255. This might me useful to implement backends not
        supporting color transparency.
        
        Returns:
            (int, int, int)
                Values of RGB channels.
        """
        
        return self._red, self._green, self._blue
    
    
    @property
    def rgba_r(self):
        """
        Gets RGBA channels tuple where each channel is defined as float in
        range 0 to 1.
        
        Returns:
            (float, float, float, float)
                Values of RGBA channels.
        """
        
        return self._red/255., self._green/255., self._blue/255., self._alpha/255.
    
    
    @property
    def rgb_r(self):
        """
        Gets RGB channels tuple where each channel is defined as float in
        range 0 to 1.
        
        Returns:
            (float, float, float)
                Values of RGB channels.
        """
        
        return self._red/255., self._green/255., self._blue/255.
    
    
    @property
    def hex(self):
        """
        Gets RGBA channels as hex string prefixed by '#'.
        
        Returns:
            str
                Hex string of #RGBA channels.
        """
        
        return "#%02x%02x%02x%02x" % (self._red, self._green, self._blue, self._alpha)
    
    
    def luminance(self):
        """
        Gets relative luminance as a value in range 0 to 1.
        
        Returns:
            float
                Relative luminance.
        """
        
        lin_r = self._red / 255.
        lin_r = lin_r / 12.92 if lin_r <= 0.04045 else math.pow(((lin_r + 0.055) / 1.055), 2.4)
        
        lin_g = self._red / 255.
        lin_g = lin_g / 12.92 if lin_g <= 0.04045 else math.pow(((lin_g + 0.055) / 1.055), 2.4)
        
        lin_b = self._red / 255.
        lin_b = lin_b / 12.92 if lin_b <= 0.04045 else math.pow(((lin_b + 0.055) / 1.055), 2.4)
        
        return 0.2126 * lin_r + 0.7152 * lin_g + 0.0722 * lin_b
    
    
    def lightness(self):
        """
        Gets perceptual lightness as a value in range 0 to 100.
        
        Returns:
            float
                Perceptual lightness.
        """
        
        lum = self.luminance()
        
        if lum <= 0.008856:
            return lum * 903.3
        
        return math.pow(lum, (1/3)) * 116 - 16
    
    
    def brightness(self):
        """
        Gets brightness as a value in range 0 to 1.
        
        Returns:
            float
                Brightness value.
        """
        
        lum = self.luminance()
        
        if lum <= 0.0031308:
            return lum * 12.92
        
        return 1.055 * math.pow(lum, 1.0/2.4) - 0.055
    
    
    def lighter(self, factor=0.2, name=None):
        """
        Creates derived color by making current color lighter. The factor
        specifies relative amount of white to be added, i.e. 1 results in full
        white color while 0 makes no change. The new color is automatically
        registered for later use if the name is specified.
        
        Args:
            factor: float
                Relative amount of white to be added in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                Lighter color.
        """
        
        # check value
        if factor < 0 or factor > 1:
            message = "Lightening factor must be in range 0 to 1! -> '%s'" % factor
            raise ValueError(message)
        
        # make color
        r = self._red + (255 - self._red) * factor
        g = self._green + (255 - self._green) * factor
        b = self._blue + (255 - self._blue) * factor
        
        return Color(r, g, b, self._alpha, name=name)
    
    
    def darker(self, factor=0.2, name=None):
        """
        Creates derived color by making current color darker. The factor
        specifies relative amount of black to be added, i.e. 1 results in full
        black color while 0 makes no change. The new color is automatically
        registered for later use if the name is specified.
        
        Args:
            factor: float
                Relative amount of black to be added in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                Darker color.
        """
        
        # check value
        if factor < 0 or factor > 1:
            message = "Darkening factor must be in range 0 to 1! -> '%s'" % factor
            raise ValueError(message)
        
        # make color
        r = self._red * (1. - factor)
        g = self._green * (1. - factor)
        b = self._blue * (1. - factor)
        
        return Color(r, g, b, self._alpha, name=name)
    
    
    def opaque(self, opacity=1, name=None):
        """
        Creates derived color by setting the opacity. 0 results in fully
        transparent color while 1 means fully opaque. The new color is
        automatically registered for later use if the name is specified.
        
        Args:
            opacity: float
                Opacity value in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                New color with specified opacity.
        """
        
        # check value
        if opacity < 0 or opacity > 1:
            message = "Opacity must be in range 0 to 1! -> '%s'" % opacity
            raise ValueError(message)
        
        # make color
        alpha = opacity * 255
        
        return Color(self._red, self._green, self._blue, alpha, name=name)
    
    
    def trans(self, transparency=1, name=None):
        """
        Creates derived color by setting the transparency. 0 results in fully
        opaque color while 1 means fully transparent. The new color is
        automatically registered for later use if the name is specified.
        
        Args:
            transparency: float
                Transparency value in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                New color with specified transparency.
        """
        
        # check value
        if transparency < 0 or transparency > 1:
            message = "Transparency must be in range 0 to 1! -> '%s'" % transparency
            raise ValueError(message)
        
        # make color
        alpha = 255 - transparency * 255
        
        return Color(self._red, self._green, self._blue, alpha, name=name)
    
    
    def invert(self, name=None):
        """
        Creates derived color by inverting all color channels.
        
        Args:
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                New inverted color.
        """
        
        return Color(255-self._red, 255-self._green, 255-self._blue, self._alpha, name=name)
    
    
    @staticmethod
    def create(value, name=None):
        """
        Creates new color from given value. The color can be specified as an
        RGB or RGBA tuple of integers, hex code, unique library name or existing
        pero.Color to get its copy. The new color is automatically registered
        for later use if the name is specified.
        
        Args:
            value: str, (int, int, int), (int, int, int, int) or pero.Color
                Any supported color definition.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                RGBA color.
        """
        
        # clone given color instance
        if isinstance(value, Color):
            return Color(value.rgba, name=name)
        
        # convert from channels
        if isinstance(value, (list, tuple)):
            return Color(value, name=name)
        
        # convert from string
        if isinstance(value, str):
            
            # convert from hex
            if value[0] == '#':
                return Color(value, name=name)
            
            # use registered name
            return Color.from_name(value)
        
        # unable to parse
        message = "Cannot create new color from given value! -> %s" % (value,)
        raise ValueError(message)
    
    
    @staticmethod
    def from_name(name):
        """
        Gets the color from library by its registered name (case in-sensitive).
        
        Args:
            name: str
                Registered color name.
        
        Returns:
            pero.Color
                RGBA color.
        """
        
        # get color
        if name in COLORS:
            return COLORS[name]
        
        # name not found
        message = "Unknown color name specified! -> '%s'" % name
        raise ValueError(message)
    
    
    @staticmethod
    def interpolate(color1, color2, x, name=None):
        """
        Creates new color by interpolating relative position between two
        colors. The new color is automatically registered for later use if the
        name is specified.
        
        Args:
            color1: str, (int, int, int), (int, int, int, int) or pero.Color
                First color.
            
            color2: str, (int, int, int), (int, int, int, int) or pero.Color
                Second color.
            
            x: float
                Relative position in %/100.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                Interpolated color.
        """
        
        # check types
        if not isinstance(color1, Color):
            color1 = Color.create(color1)
        
        if not isinstance(color2, Color):
            color2 = Color.create(color2)
        
        # interpolate channels
        x = float(x)
        red = color1.red + x*(color2.red-color1.red)
        green = color1.green + x*(color2.green-color1.green)
        blue = color1.blue + x*(color2.blue-color1.blue)
        alpha = color1.alpha + x*(color2.alpha-color1.alpha)
        
        # crop channels
        red = max(0., min(255., red))
        green = max(0., min(255., green))
        blue = max(0., min(255., blue))
        alpha = max(0., min(255., alpha))
        
        # make color
        return Color(red, green, blue, alpha, name=name)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import inspect
from .. events import EvtHandler
from . undefined import UNDEF
from . prop import Property

# define property names splitter
PROP_SPLITTER = '_'

# init properties cache
_PROPERTIES_CACHE = {}


class Include(object):
    """
    This class is used as a tool to include all available properties of another
    property set. The properties can be included with optional prefix, which is
    added then to the name of each included property, and custom defaults
    provided as overrides. In addition, the 'dynamic' and 'nullable' flags can
    also be changed for all the included property.
    """
    
    
    def __init__(self, prop_set, prefix="", dynamic=None, nullable=None, exclude=None, **overrides):
        """
        Initializes a new instance of Include.
        
        Args:
            prop_set: pero.PropertySet class
                Property set class from which to include all the properties.
            
            prefix: str
                Optional prefix to be added to the names of all included
                properties.
            
            dynamic: bool or None
                If set to True or False the value overwrites the original value
                of the all included properties.
            
            nullable: bool or None
                If set to True or False the value overwrites the original value
                of the all included properties.
            
            exclude: (str,)
                Names of the properties to ignore. Original names without the
                prefix must be used.
            
            overrides: str:any pairs
                Overwrites for default values of specific properties. Original
                names without the prefix must be used.
        """
        
        # check type
        if not issubclass(prop_set, PropertySet):
            message = "Properties must be subclass of pero.PropertySet! -> %s" % type(prop_set)
            raise TypeError(message)
        
        # set values
        self._property_set = prop_set
        self._prefix = prefix
        self._dynamic = dynamic
        self._nullable = nullable
        self._exclude = set(exclude) if exclude else set()
        self._overrides = overrides
    
    
    def make_properties(self):
        """
        Creates clones of all properties to be included and adjusts each name
        by provided prefix.
        
        Returns:
            (pero.Property,)
                List of initialized properties.
        """
        
        properties = []
        
        # clone properties
        for prop in self._property_set.properties():
            
            # skip if excluded
            if prop.name in self._exclude:
                continue
            
            # get name
            name = prop.name
            if self._prefix:
                name = self._prefix + name
            
            # get default
            default = prop.default
            if prop.name in self._overrides:
                default = self._overrides[prop.name]
            
            # get dynamic
            dynamic = prop.dynamic
            if self._dynamic is not None:
                dynamic = self._dynamic
            
            # get nullable
            nullable = prop.nullable
            if self._nullable is not None:
                nullable = self._nullable
            
            # add cloned property
            properties.append(prop.clone(
                name = name,
                default = default,
                dynamic = dynamic,
                nullable = nullable))
        
        return properties


class PropertySetMeta(type):
    """Defines a metaclass for pero.PropertySet classes."""
    
    
    def __new__(cls, cls_name, bases, cls_dict):
        """Creates a new parent class."""
        
        # finalize properties
        for name, item in tuple(cls_dict.items()):
            
            # set property name
            if isinstance(item, Property):
                item._name = name
            
            # insert properties from Includes
            elif isinstance(item, Include):
                
                # remove include itself
                del cls_dict[name]
                
                # insert contained properties
                for prop in item.make_properties():
                    if prop.name in cls_dict:
                        message = "Property with the name '%s' is already present!" % prop.name
                        raise KeyError(message)
                    cls_dict[prop.name] = prop
        
        # init new class
        return type.__new__(cls, cls_name, bases, cls_dict)


class PropertySet(EvtHandler, metaclass=PropertySetMeta):
    """
    Abstract base class for all property-having classes.
    
    This class also serves as an event handler, which means it can fire events
    using the 'fire' method and specific listeners can be attached to it. By
    default only the pero.EVT_PROPERTY_CHANGED is fired every time a
    property is changed.
    """
    
    
    def __init__(self, **overrides):
        """
        Initializes a new instance of PropertySet.
        
        Args:
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        super().__init__()
        
        # get available properties
        self._properties = {p.name: p for p in self.properties()}
        self._locked = set()
        self._held = set()
        
        # set given properties
        self.set_properties(overrides, True)
    
    
    def __call__(self, **overrides):
        """
        Updates specified properties.
        
        Args:
            overrides: str:any pairs
                Value overwrites for specific properties.
        """
        
        # set given properties
        self.set_properties(overrides, True)
    
    
    def has_property(self, name):
        """
        Checks whether specified property exists.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property.
        
        Returns:
            bool
                True if the property exists, False otherwise.
        """
        
        # check known property
        if name in self._properties:
            return True
        
        # check child property
        path = self.get_property_path(name, False)
        return path is not None
    
    
    def get_property(self, name, source=UNDEF, overrides=None, native=False):
        """
        Gets the value of specified property.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        If specified property is available within 'overrides' the value in
        'overrides' is used instead of current one. If allowed and the value is
        callable but still not of the requested type, given 'source' is provided
        as the argument for calling it and returned value is finally used.
        
        In some cases it might be useful to retrieve the callable function
        itself, without applying it onto given source. This can be achieved by
        setting the 'native' argument to True.
        
        Args:
            name: str
                Name of the property to be retrieved.
            
            source: any
                Data source to be used for retrieving the final value of
                callable properties.
            
            overrides: dict or None
                Highest priority properties to be used instead of current value.
            
            native: bool
                If set to True callable properties are returned directly
                without calling and using the source.
        
        Returns:
            any
                Property value.
        """
        
        # get property path
        parent, prop = self.get_property_path(name, True)[-2:]
        
        # get value
        if overrides and name in overrides:
            value = prop.parse(overrides[name])
        else:
            value = getattr(parent, prop.name)
        
        # use native value
        if native:
            return value
        
        # requested type already
        if isinstance(value, prop.types):
            return value
        
        # dynamic property
        if callable(value):
            value = value(source)
        
        # parse value
        return prop.parse(value)
    
    
    def get_property_path(self, name, raise_error=False):
        """
        Gets the full path of specified property. It includes all the parents
        and finally the property itself.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property.
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be retrieved.
        
        Returns:
            (..., pero.PropertySet, pero.Property) or None
                Property path or None if not found.
        """
        
        # init path
        path = [self]
        
        # get known property
        if name in self._properties:
            return path + [self._properties[name]]
        
        # get first split
        idx = name.find(PROP_SPLITTER)
        
        # search child properties
        while idx > 0:
            
            # get parent name
            parent = name[:idx]
            
            # check parent
            if parent in self._properties:
                parent_prop = getattr(self, parent)
                if isinstance(parent_prop, PropertySet):
                    child_prop = parent_prop.get_property_path(name[idx+1:])
                    if child_prop:
                        return path + child_prop
            
            # get next split
            idx = name.find(PROP_SPLITTER, idx+1)
        
        # raise error for unknown property
        if raise_error:
            message = "%s has no property '%s'!" % (self.__class__.__name__, name)
            raise AttributeError(message)
        
        # property not found
        return None
    
    
    def get_own_overrides(self, overrides):
        """
        Extracts the overrides for direct properties of current property set.
        
        Args:
            overrides: dict or None
                Overrides to extract the properties from.
        
        Returns:
            dict
                Extracted self overrides as {name: value}.
        """
        
        # check overrides
        if overrides is None:
            return {}
        
        # init own overrides
        own_overrides = {}
        
        # get own overrides
        for name in overrides:
            if name in self._properties:
                own_overrides[name] = overrides[name]
        
        return own_overrides
    
    
    def get_child_overrides(self, child_name, overrides, skip=None, loose=True):
        """
        Extracts the overrides for child property set. A property is considered
        as child-related if it starts with specified 'child_name' followed by
        pero.PROP_SPLITTER character and it is not a direct property of current
        property set.
        
        Args:
            child_name: str
                Name of the child property set.
            
            overrides: dict or None
                Overrides to extract the properties from.
            
            skip: (str,)
                Collection of properties to skip.
            
            loose: bool
                If set to True, matching properties existing also in current set
                will be included as well.
        
        Returns:
            dict
                Extracted child overrides as {name: value}.
        """
        
        # check overrides
        if overrides is None:
            return {}
        
        # get prefix
        prefix = child_name

        # add splitter
        if prefix and prefix[-1] != PROP_SPLITTER:
            prefix += PROP_SPLITTER
        
        # init child overrides
        child_overrides = {}
        
        # get child overrides
        for name in overrides:
            
            # skip property
            if skip and name in skip:
                continue
            
            # skip current
            if not loose and name in self._properties:
                continue
            
            # skip without prefix
            if not name.startswith(prefix) or name == prefix:
                continue
            
            # add override
            new_name = name.split(prefix, 1)[1]
            child_overrides[new_name] = overrides[name]
        
        return child_overrides
    
    
    def set_property(self, name, value, raise_error=True):
        """
        Sets given value for specified property.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property to be set.
            
            value: any
                Property value to be set.
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be set.
        """
        
        # set known property
        if name in self._properties:
            setattr(self, name, value)
            return
        
        # get child property path
        path = self.get_property_path(name, raise_error)
        if not path:
            return
        
        # set value
        setattr(path[-2], path[-1].name, value)
    
    
    def set_properties(self, properties, raise_error=True):
        """
        Sets multiple properties in a batch using name:value dictionary.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            properties: dict
                Properties names and values to be set.
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be set.
        """
        
        for name, value in sorted(properties.items()):
            self.set_property(name, value, raise_error)
    
    
    def set_properties_from(self, prop_set, src_prefix="", dst_prefix="", source=UNDEF, overrides=None, skip=None, native=False):
        """
        Sets values of all shared properties from given property set to current
        property set.
        
        If any property is available within 'overrides' the value in
        'overrides' is used instead of current one. If allowed and the value is
        callable but still not of the requested type, given 'source' is provided
        as argument for calling it and returned value is finally used.
        
        In some cases it might be useful to retrieve the callable function
        itself, without applying it onto given source. This can be achieved by
        setting the 'native' argument to True.
        
        Args:
            prop_set: pero.PropertySet
                Property set from which to retrieve the properties.
            
            src_prefix: str
                Prefix used for shared properties in the source set. Shared
                properties without this prefix will be skipped.
            
            dst_prefix: str
                Prefix used for shared properties in the destination set. Shared
                properties without this prefix will be skipped.
            
            source: any
                Data source to be used for retrieving the final value of
                callable properties.
            
            overrides: dict or None
                Highest priority properties to be used instead of current values.
                The names should include the 'src_prefix'.
            
            skip: (str,)
                Collection of properties to skip. The names should include
                the 'src_prefix'.
            
            native: bool
                If set to True callable properties from the source are used
                directly without calling.
        """
        
        # add splitter
        if src_prefix and src_prefix[-1] != PROP_SPLITTER:
            src_prefix += PROP_SPLITTER
        
        if dst_prefix and dst_prefix[-1] != PROP_SPLITTER:
            dst_prefix += PROP_SPLITTER
        
        # process source properties
        for prop in prop_set.properties():
            
            # get source name without prefix
            name = prop.name
            if src_prefix and name.startswith(src_prefix):
                name = name[len(src_prefix):]
            elif src_prefix:
                continue
            
            # finalize names
            src_name = prop.name
            dst_name = dst_prefix + name
            
            # skip property
            if skip and src_name in skip:
                continue
            
            # set shared properties
            if dst_name in self._properties:
                value = prop_set.get_property(src_name, source, overrides, native)
                setattr(self, dst_name, value)
    
    
    def set_properties_to(self, prop_set, src_prefix="", dst_prefix="", source=UNDEF, overrides=None, skip=None, native=False):
        """
        Sets values of all shared properties from current property set to given
        property set.
        
        If any property is available within 'overrides' the value in
        'overrides' is used instead of current one. If allowed and the value is
        callable but still not of the requested type, given 'source' is provided
        as argument for calling it and returned value is finally used.
        
        In some cases it might be useful to retrieve the callable function
        itself, without applying it onto given source. This can be achieved by
        setting the 'native' argument to True.
        
        Args:
            prop_set: pero.PropertySet
                Property set to which set the properties.
            
            src_prefix: str
                Prefix used for shared properties in the source set. Shared
                properties without this prefix will be skipped.
            
            dst_prefix: str
                Prefix used for shared properties in the destination set. Shared
                properties without this prefix will be skipped.
            
            source: any
                Data source to be used for retrieving the final value of
                callable properties.
            
            overrides: dict or None
                Highest priority properties to be used instead of current values.
                The names should include the 'src_prefix'.
            
            skip: (str,)
                Collection of properties to skip. The names should include
                the 'src_prefix'.
            
            native: bool
                If set to True callable properties from the source are used
                directly without calling.
        """
        
        # add splitter
        if src_prefix and src_prefix[-1] != PROP_SPLITTER:
            src_prefix += PROP_SPLITTER
        
        if dst_prefix and dst_prefix[-1] != PROP_SPLITTER:
            dst_prefix += PROP_SPLITTER
        
        # process current properties
        for prop in self.properties():
            
            # get current name without prefix
            name = prop.name
            if src_prefix and name.startswith(src_prefix):
                name = name[len(src_prefix):]
            elif src_prefix:
                continue
            
            # finalize names
            src_name = prop.name
            dst_name = dst_prefix + name
            
            # skip property
            if skip and src_name in skip:
                continue
            
            # set shared properties
            if dst_name in prop_set._properties:
                value = self.get_property(src_name, source, overrides, native)
                setattr(prop_set, dst_name, value)
    
    
    def lock_property(self, name, lock=True, raise_error=True):
        """
        Locks or unlocks specified property to disable or enable any further
        changes.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property to be locked/unlocked.
            
            lock: bool
                Specifies whether the property should be locked (True) or
                unlocked (False).
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be locked/unlocked.
        """
        
        # lock/unlock known property
        if name in self._properties:
            if lock:
                self._locked.add(name)
            else:
                self._locked.discard(name)
            return
        
        # get child property path
        path = self.get_property_path(name, raise_error)
        if not path:
            return
        
        # lock/unlock
        path[-2].lock_property(path[-1].name, lock)
    
    
    def hold_property(self, name, hold=True, raise_error=True):
        """
        Holds or releases specified property to keep its previous value if the
        new value is undefined (pero.UNDEF).
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property to be held/released.
            
            hold: bool
                Specifies whether the property should be held (True) or
                released (False).
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be held/released.
        """
        
        # hold/release known property
        if name in self._properties:
            if hold:
                self._held.add(name)
            else:
                self._held.discard(name)
            return
        
        # get child property path
        path = self.get_property_path(name, raise_error)
        if not path:
            return
        
        # hold/release
        path[-2].hold_property(path[-1].name, hold)
    
    
    def is_property_locked(self, name, raise_error=True):
        """
        Checks whether specified property is locked.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property.
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be checked.
        
        Returns:
            bool
                True if the property is locked, False otherwise.
        """
        
        # check known property
        if name in self._properties:
            return name in self._locked
        
        # get child property path
        path = self.get_property_path(name, raise_error)
        if not path:
            return
        
        # check property
        path[-2].is_property_locked(path[-1].name)
    
    
    def is_property_held(self, name, raise_error=True):
        """
        Checks whether specified property is held.
        
        If specified property is not found directly, child properties are
        searched automatically considering the pero.PROP_SPLITTER character as
        a splitter between parent and child property name. This mechanism
        assumes the parent is derived from pero.PropertySet and it must be
        initialized already.
        
        Args:
            name: str
                Name of the property.
            
            raise_error: bool
                If set to True, an error is raised if unknown property is about
                to be checked.
        
        Returns:
            bool
                True if the property is held, False otherwise.
        """
        
        # check known property
        if name in self._properties:
            return name in self._held
        
        # get child property path
        path = self.get_property_path(name, raise_error)
        if not path:
            return
        
        # check property
        path[-2].is_property_held(path[-1].name)
    
    
    def clone(self, source=UNDEF, overrides=None, native=False, deep=False):
        """
        Creates a clone of current instance. A new pero.PropertySet
        is created with all the properties cloned. Depending on the 'deep'
        value, the actual values of the properties are just copied, but not
        cloned (False) or cloned as well, if they are also instances of the
        PropertySet (True).
        
        If any property is available within 'overrides' the value in
        'overrides' is used instead of current one. If allowed and the value is
        callable but still not of the requested type, given 'source' is provided
        as argument for calling it and returned value is finally used. This is
        very useful for defining a 'template' with some properties as functions
        and create a cloned instance with final values. A nice example could be
        labels for scatter plot.
        
        In some cases it might be useful to retrieve the callable function
        itself, without applying it onto given source. This can be achieved by
        setting the 'native' argument to True.
        
        Args:
            source: any
                Data source to be used for retrieving the final value of
                callable properties.
            
            overrides: dict or None
                Highest priority properties to be used instead of current values.
            
            native: bool
                If set to True callable properties are returned directly
                without calling.
            
            deep: bool
                If set to True properties, which are instances of PropertySet
                will be cloned as well.
        
        Returns:
            pero.PropertySet
                Cloned property set.
        """
        
        # init clone instance
        clone = self.__class__()
        
        # get properties
        for name in self._properties:
            
            # get value
            value = self.get_property(name, source, overrides, native)
            
            # clone value
            if deep and isinstance(value, PropertySet):
                child_overrides = self.get_child_overrides(name, overrides, loose=False)
                value = value.clone(source, child_overrides, native, deep)
            
            # set value
            setattr(clone, name, value)
        
        # keep locks
        clone._locked = set(self._locked)
        
        # keep holds
        clone._held = set(self._held)
        
        return clone
    
    
    @classmethod
    def properties(cls):
        """
        Gets all available properties of the class.
        
        Returns:
            (pero.Property,)
                Available properties.
        """
        
        key = cls
        
        # use cache
        if key in _PROPERTIES_CACHE:
            return _PROPERTIES_CACHE[key]
        
        # collect properties
        properties = []
        for item in inspect.getmembers(key):
            if isinstance(item[1], Property):
                properties.append(item[1])
        
        # update cache
        _PROPERTIES_CACHE[key] = tuple(properties)
        
        return properties



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import UNDEF, PropertySet, SetProperty


class Tool(PropertySet):
    """Abstract base class for interactivity tools."""
    
    keys = SetProperty(UNDEF, dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Tool."""
        
        super().__init__(**overrides)
        
        # init buffers
        if self.keys == UNDEF:
            self.keys = set()
    
    
    def on_size(self, evt):
        """
        This method should be overridden to provide specific handling
        of window size-change event.
        
        Args:
            evt: pero.SizeEvt
                Event to process.
        """
        
        pass
    
    
    def on_key_down(self, evt):
        """
        This method should be overridden to provide specific handling
        of key-down event.
        
        Args:
            evt: pero.KeyDownEvt
                Event to process.
        """
        
        self.add_key(evt.key)
    
    
    def on_key_up(self, evt):
        """
        This method should be overridden to provide specific handling
        of key-up event.
        
        Args:
            evt: pero.KeyUpEvt
                Event to process.
        """
        
        self.remove_key(evt.key)
    
    
    def on_mouse_enter(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-enter event.
        
        Args:
            evt: pero.MouseEnterEvt
                Event to process.
        """
        
        pass
    
    
    def on_mouse_leave(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-leave event.
        
        Args:
            evt: pero.MouseLeaveEvt
                Event to process.
        """
        
        self.clear_keys()
    
    
    def on_mouse_motion(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-motion event.
        
        Args:
            evt: pero.MouseMotionEvt
                Event to process.
        """
        
        pass
    
    
    def on_mouse_scroll(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-scroll event.
        
        Args:
            evt: pero.MouseScrollEvt
                Event to process.
        """
        
        pass
    
    
    def on_mouse_down(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-button-down event.
        
        Args:
            evt: pero.LeftDownEvt, pero.RightDownEvt or pero.MiddleDownEvt
                Event to process.
        """
        
        pass
    
    
    def on_mouse_up(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-button-up event.
        
        Args:
            evt: pero.LeftUpEvt, pero.RightUpEvt or pero.MiddleUpEvt
                Event to process.
        """
        
        pass
    
    
    def on_mouse_dclick(self, evt):
        """
        This method should be overridden to provide specific handling
        of mouse-button-double-click event.
        
        Args:
            evt: pero.LeftDClickEvt, pero.RightDClickEvt or pero.MiddleDClickEvt
                Event to process.
        """
        
        pass
    
    
    def on_touch_start(self, evt):
        """
        This method should be overridden to provide specific handling
        of touch-start event.
        
        Args:
            evt: pero.TouchStartEvt
                Event to process.
        """
        
        pass
    
    
    def on_touch_end(self, evt):
        """
        This method should be overridden to provide specific handling
        of touch-end event.
        
        Args:
            evt: pero.TouchEndEvt
                Event to process.
        """
        
        pass
    
    
    def on_touch_move(self, evt):
        """
        This method should be overridden to provide specific handling
        of touch-move event.
        
        Args:
            evt: pero.TouchMoveEvt
                Event to process.
        """
        
        pass
    
    
    def on_touch_cancel(self, evt):
        """
        This method should be overridden to provide specific handling
        of touch-cancel event.
        
        Args:
            evt: pero.TouchCancelEvt
                Event to process.
        """
        
        pass
    
    
    def on_touch_dtap(self, evt):
        """
        This method should be overridden to provide specific handling
        of touch-double-tap event.
        
        Args:
            evt: pero.TouchDTapEvt
                Event to process.
        """
        
        pass
    
    
    def add_key(self, key):
        """
        Remembers given key.
        
        Args:
            key: pero.KEY
                A key to remember as any value from the pero.KEY enum.
        """
        
        self.keys.add(key)
    
    
    def remove_key(self, key):
        """
        Removes given key.
        
        Args:
            key: pero.KEY
                A key to remove as any value from the pero.KEY enum.
        """
        
        self.keys.discard(key)
    
    
    def clear_keys(self):
        """Removes all keys."""
    
        self.keys.clear()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . loader import Qt, QFont
from ... enums import *


QT_LINE_CAP = {
    LINE_CAP_BUTT: Qt.PenCapStyle.FlatCap,
    LINE_CAP_SQUARE: Qt.PenCapStyle.SquareCap,
    LINE_CAP_ROUND: Qt.PenCapStyle.RoundCap}

QT_LINE_JOIN = {
    LINE_JOIN_BEVEL: Qt.PenJoinStyle.BevelJoin,
    LINE_JOIN_MITER: Qt.PenJoinStyle.MiterJoin,
    LINE_JOIN_ROUND: Qt.PenJoinStyle.RoundJoin}

QT_LINE_STYLE = {
    LINE_STYLE_CUSTOM: Qt.PenStyle.CustomDashLine,
    LINE_STYLE_SOLID: Qt.PenStyle.SolidLine,
    LINE_STYLE_DOTTED: DASH_VALUES_DOTTED,
    LINE_STYLE_DASHED: DASH_VALUES_DASHED,
    LINE_STYLE_DASHDOTTED: DASH_VALUES_DASHDOTTED}

QT_FILL_STYLE = {
    FILL_STYLE_TRANS: Qt.BrushStyle.NoBrush,
    FILL_STYLE_SOLID: Qt.BrushStyle.SolidPattern}

QT_FILL_RULE = {
    FILL_RULE_EVENODD: Qt.FillRule.OddEvenFill,
    FILL_RULE_WINDING: Qt.FillRule.WindingFill}

QT_FONT_FAMILY = {
    FONT_FAMILY_SERIF: FONT_FACE_SERIF,
    FONT_FAMILY_SANS: FONT_FACE_SANS,
    FONT_FAMILY_MONO: FONT_FACE_MONO}

QT_FONT_STYLE = {
    FONT_STYLE_NORMAL: QFont.Style.StyleNormal,
    FONT_STYLE_ITALIC: QFont.Style.StyleItalic}

QT_FONT_WEIGHT = {
    FONT_WEIGHT_NORMAL: QFont.Weight.Normal,
    FONT_WEIGHT_LIGHT: QFont.Weight.Light,
    FONT_WEIGHT_BOLD: QFont.Weight.Bold,
    FONT_WEIGHT_BLACK: QFont.Weight.Black,
    FONT_WEIGHT_HEAVY: QFont.Weight.ExtraBold,
    FONT_WEIGHT_SEMIBOLD: QFont.Weight.DemiBold,
    FONT_WEIGHT_MEDIUM: QFont.Weight.Medium,
    FONT_WEIGHT_ULTRALIGHT: QFont.Weight.ExtraLight,
    FONT_WEIGHT_THIN: QFont.Weight.Thin}

QT_CURSORS = {
    CURSOR_BLANK: Qt.CursorShape.BlankCursor,
    CURSOR_ARROW: Qt.CursorShape.ArrowCursor,
    CURSOR_HAND: Qt.CursorShape.OpenHandCursor,
    CURSOR_SIZEWE: Qt.CursorShape.SizeHorCursor,
    CURSOR_SIZENS: Qt.CursorShape.SizeVerCursor}

QT_KEYS = {
    Qt.Key.Key_Backspace: KEY_BACK,
    Qt.Key.Key_Tab: KEY_TAB,
    Qt.Key.Key_Return: KEY_RETURN,
    
    Qt.Key.Key_Shift: KEY_SHIFT,
    Qt.Key.Key_Control: KEY_CTRL,
    Qt.Key.Key_Alt: KEY_ALT,
    
    Qt.Key.Key_Escape: KEY_ESC,
    Qt.Key.Key_Space: KEY_SPACE,
    
    Qt.Key.Key_PageUp: KEY_PAGEUP,
    Qt.Key.Key_PageDown: KEY_PAGEDOWN,
    Qt.Key.Key_End: KEY_END,
    Qt.Key.Key_Home: KEY_HOME,
    Qt.Key.Key_Left: KEY_LEFT,
    Qt.Key.Key_Up: KEY_UP,
    Qt.Key.Key_Right: KEY_RIGHT,
    Qt.Key.Key_Down: KEY_DOWN,
    
    Qt.Key.Key_Delete: KEY_DELETE,
    
    Qt.Key.Key_0: KEY_0,
    Qt.Key.Key_1: KEY_1,
    Qt.Key.Key_2: KEY_2,
    Qt.Key.Key_3: KEY_3,
    Qt.Key.Key_4: KEY_4,
    Qt.Key.Key_5: KEY_5,
    Qt.Key.Key_6: KEY_6,
    Qt.Key.Key_7: KEY_7,
    Qt.Key.Key_8: KEY_8,
    Qt.Key.Key_9: KEY_9,
    
    Qt.Key.Key_A: KEY_A,
    Qt.Key.Key_B: KEY_B,
    Qt.Key.Key_C: KEY_C,
    Qt.Key.Key_D: KEY_D,
    Qt.Key.Key_E: KEY_E,
    Qt.Key.Key_F: KEY_F,
    Qt.Key.Key_G: KEY_G,
    Qt.Key.Key_H: KEY_H,
    Qt.Key.Key_I: KEY_I,
    Qt.Key.Key_J: KEY_J,
    Qt.Key.Key_K: KEY_K,
    Qt.Key.Key_L: KEY_L,
    Qt.Key.Key_M: KEY_M,
    Qt.Key.Key_N: KEY_N,
    Qt.Key.Key_O: KEY_O,
    Qt.Key.Key_P: KEY_P,
    Qt.Key.Key_Q: KEY_Q,
    Qt.Key.Key_R: KEY_R,
    Qt.Key.Key_S: KEY_S,
    Qt.Key.Key_T: KEY_T,
    Qt.Key.Key_U: KEY_U,
    Qt.Key.Key_V: KEY_V,
    Qt.Key.Key_W: KEY_W,
    Qt.Key.Key_X: KEY_X,
    Qt.Key.Key_Y: KEY_Y,
    Qt.Key.Key_Z: KEY_Z,
    
    Qt.Key.Key_F1: KEY_F1,
    Qt.Key.Key_F2: KEY_F2,
    Qt.Key.Key_F3: KEY_F3,
    Qt.Key.Key_F4: KEY_F4,
    Qt.Key.Key_F5: KEY_F5,
    Qt.Key.Key_F6: KEY_F6,
    Qt.Key.Key_F7: KEY_F7,
    Qt.Key.Key_F8: KEY_F8,
    Qt.Key.Key_F9: KEY_F9,
    Qt.Key.Key_F10: KEY_F10,
    Qt.Key.Key_F11: KEY_F11,
    Qt.Key.Key_F12: KEY_F12}

QT_TOUCH_STATE = {
    Qt.TouchPointState.TouchPointPressed: TOUCH_PRESSED,
    Qt.TouchPointState.TouchPointReleased: TOUCH_RELEASED,
    Qt.TouchPointState.TouchPointMoved: TOUCH_MOVED,
    Qt.TouchPointState.TouchPointStationary: TOUCH_STATIC}

QT_RASTER_TYPES = {
    '.bmp',
    '.gif',
    '.jpg',
    '.jpeg',
    '.png'}

QT_VECTOR_TYPES = {
    '.pdf'}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . proxy import Proxy
from . event import Event


class EvtHandler(object):
    """
    This class represents an event raising base class, to which specific
    callbacks can be attached to be called when appropriate event is fired.
    When a specific event is fired, bound callbacks are called in reversed
    order, so the last added callback will be called first. Calling of
    registered callbacks continues until all are called or until one of them
    cancels the event by calling the 'cancel' method.
    """
    
    
    def __init__(self):
        """Initializes a new instance of EvtHandler."""
        
        self._callbacks = {}
    
    
    def bind(self, evt_type, callback, **kwargs):
        """
        Registers given callback for specific event. Additional keyword
        arguments can be specified to be always provided when the event is
        fired. These arguments will be used as defaults but they can be
        overridden by the same argument provided directly to the 'fire' method.
        
        Args:
            evt_type: str or pero.Event
                Specific event type to which the callback should be attached.
            
            callback: callable
                Callback to be called when event is fired.
            
            kwargs: str:any pairs
                Additional keyword arguments.
        """
        
        # get event type
        evt_type = self._get_evt_type(evt_type)
        
        # register event
        if evt_type not in self._callbacks:
            self._callbacks[evt_type] = []
        
        # init proxy
        proxy = Proxy(callback)
        
        # add callback
        self._callbacks[evt_type].append((proxy, kwargs))
    
    
    def unbind(self, evt_type, callback, **kwargs):
        """
        Removes given callback if registered for specified event. If additional
        keyword arguments are provided, they must match exactly all the
        arguments provided when the callback was attached. If not provided,
        just the callback must match.
        
        Args:
            evt_type: str or pero.Event
                Specific event type to which the callback was attached.
            
            callback: callable
                Callback to be removed.
            
            kwargs: str:any pairs
                Additional keyword arguments.
        
        Returns:
            bool
                Returns True if any callback was removed, False otherwise.
        """
        
        removed = False
        
        # get event type
        evt_type = self._get_evt_type(evt_type)
        
        # get callbacks
        callbacks = self._callbacks.get(evt_type, None)
        if not callbacks:
            return removed
        
        # init proxy
        proxy = Proxy(callback)
        
        # remove matching callback
        for item in callbacks[:]:
            
            if item[0] != proxy:
                continue
            
            if kwargs and item[0] != kwargs:
                continue
            
            callbacks.remove(item)
            removed = True
        
        return removed
    
    
    def fire(self, evt, **kwargs):
        """
        Fires given event by calling all registered callbacks with given params.
        When a specific event is fired, bound callbacks are called in reversed
        order, so the last added callback will be called first. Calling of
        registered callbacks continues until all are called or until
        one of them cancels the event by calling 'cancel' method.
        
        Args:
            evt: pero.Event
                Event instance which should be processed.
            
            kwargs: str:any pairs
                Additional keyword arguments.
        """
        
        # get callbacks
        callbacks = self._callbacks.get(evt.TYPE, None)
        if not callbacks:
            return
        
        # call callbacks
        for callback, params in reversed(callbacks):
            
            # get params
            params = dict(params, **kwargs)
            
            # call callback
            callback(evt, **params)
            
            # check if canceled
            if evt.is_canceled():
                return
    
    
    def _get_evt_type(self, evt):
        """Gets event type."""
        
        if type(evt) == str:
            return evt
        
        if isinstance(evt, Event):
            return evt.TYPE
        
        if issubclass(evt, Event):
            return evt.TYPE



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from ... drawing import Graphics
from . canvas import JsonCanvas


class Image(JsonCanvas, Graphics):
    """
    Image represents a combination of pero.JsonCanvas and pero.Graphics. This
    provides access to all the main drawing methods with a possibility to be
    later drawn into any standard canvas. In addition, two convenient methods
    are available as shortcuts to 'export' or 'show' the image using available
    default drawing backend or viewer, depending on requested format.
    """
    
    
    def show(self, title=None, width=None, height=None, backend=None, **options):
        """
        Shows the image in available viewer app. This method makes sure
        appropriate backend canvas is created and provided to the 'draw' method.
        
        Note that is just a convenient scripting shortcut and this method cannot
        be used if the image is already part of any UI app.
        
        Args:
            title: str or None
                Viewer frame title.
            
            width: float or None
                Viewer width in device units. If set to None, current image
                width is used.
            
            height: float or None
                Viewer height in device units. If set to None, current image
                height is used.
            
            backend: pero.BACKEND
                Specific backend to be used. The value must be an item from the
                pero.BACKEND enum.
            
            options: str:any pairs
                Additional parameters for specific backend.
        """
        
        # get size
        if width is None:
            width = self.width
        if height is None:
            height = self.height
        
        # show image
        super().show(title, width, height, backend, **options)
    
    
    def export(self, path, width=None, height=None, backend=None, **options):
        """
        Draws the image into specified file using the format determined
        automatically from the file extension. This method makes sure
        appropriate backend canvas is created and provided to the 'draw' method.
        
        Note that is just a convenient scripting shortcut and this method cannot
        be used if the image is already part of any UI app.
        
        Args:
            path: str
                Full path of a file to save the image into.
            
            width: float or None
                Image width in device units. If set to None, current image width
                is used.
            
            height: float or None
                Image height in device units. If set to None, current image
                height is used.
            
            backend: pero.BACKEND
                Specific backend to be used. The value must be an item from the
                pero.BACKEND enum.
            
            options: str:any pairs
                Additional parameters for specific backend.
        """
        
        # get size
        if width is None:
            width = self.width
        if height is None:
            height = self.height
        
        # export image
        super().export(path, width, height, backend, **options)
    
    
    def draw(self, canvas, *args, **kwargs):
        """
        Uses given canvas to draw the image.
        
        Args:
            canvas: pero.Canvas
                Canvas to be used for rendering.
        """
        
        canvas.draw_json(self.get_json())
    
    
    @staticmethod
    def from_json(dump):
        """
        Creates a new pero.Image from given JSON dump.
        
        Args:
            dump: str or dict
                Image JSON dump.
        
        Returns:
            pero.Image
        """
        
        img = Image()
        img.draw_json(dump)
        
        return img
    
    
    @staticmethod
    def load_json(path):
        """
        Creates a new pero.Image from given JSON dump path.
        
        Args:
            path: str or dict
                Path to JSON dump.
        
        Returns:
            pero.Image
        """
        
        with open(path) as dump:
            return Image.from_json(dump.read())



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for levels scale."""
    
    
    def test_scale_thresholds(self):
        """Tests whether scale works correctly for exact thresholds."""
        
        levels = ["red", "green", "blue"]
        thresholds = [1, 2, 3]
        
        scale = pero.LevelScale(
            in_range = thresholds,
            out_range = levels)
        
        self.assertEqual(scale.scale(1), "red")
        self.assertEqual(scale.scale(2), "green")
        self.assertEqual(scale.scale(3), "blue")
    
    
    def test_scale_inside(self):
        """Tests whether scale works correctly for values inside."""
        
        levels = ["red", "green", "blue"]
        thresholds = [1, 2, 3]
        
        scale = pero.LevelScale(
            in_range = thresholds,
            out_range = levels)
        
        self.assertEqual(scale.scale(1.5), "green")
        self.assertEqual(scale.scale(2.5), "blue")
    
    
    def test_scale_outside(self):
        """Tests whether scale works correctly for values outside."""
        
        levels = ["red", "green", "blue"]
        thresholds = [1, 2, 3]
        
        scale = pero.LevelScale(
            in_range = thresholds,
            out_range = levels)
        
        self.assertEqual(scale.scale(.5), "red")
        self.assertEqual(scale.scale(3.5), None)
    
    
    def test_invert(self):
        """Tests whether invert works correctly."""
        
        levels = ["red", "green", "blue"]
        thresholds = [1, 2, 3]
        
        scale = pero.LevelScale(
            in_range = thresholds,
            out_range = levels)
        
        self.assertEqual(scale.invert("red"), 1)
        self.assertEqual(scale.invert("green"), 2)
        self.assertEqual(scale.invert("blue"), 3)
        self.assertEqual(scale.invert("black"), None)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . color import Color, COLORS
from . palette import Palette, PALETTES
from . gradient import Gradient, GRADIENTS

# register named objects
from . named import *



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for linear ticker."""
    
    
    def test_major_ticks(self):
        """Tests whether major ticks are calculated correctly."""
        
        ticker = pero.LinTicker(major_splits=(5,3,2))
        
        # test ascending
        ticker(start=0, end=10, major_count=3)
        ticks = ticker.major_ticks()
        
        model = (0, 3, 6, 9)
        self.assertEqual(ticks, model)
        
        ticker(start=0, end=.1, major_count=5)
        ticks = ticker.major_ticks()
        
        model = (0, .02, .04, .06, .08, .1)
        self.assertEqual(ticks, model)
        
        # test descending
        ticker(start=10, end=0, major_count=3)
        ticks = ticker.major_ticks()
        
        model = (9, 6, 3, 0)
        self.assertEqual(ticks, model)
    
    
    def test_minor_ticks(self):
        """Tests whether minor ticks are calculated correctly."""
        
        ticker = pero.LinTicker(major_splits=(5,3,2))
        
        # test ascending
        ticker(start=0, end=10, major_count=3, minor_count=5)
        ticks = ticker.minor_ticks()
        
        model = (0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 5, 5.5, 6, 6.5, 7, 7.5, 8, 8.5, 9, 9.5, 10)
        self.assertEqual(ticks, model)
        
        # test descending
        ticker(start=10, end=0, major_count=3, minor_count=5)
        ticks = ticker.minor_ticks()
        
        model = (10.0, 9.5, 9.0, 8.5, 8.0, 7.5, 7.0, 6.5, 6.0, 5.5, 5.0, 4.5, 4.0, 3.5, 3.0, 2.5, 2.0, 1.5, 1.0, 0.5, 0.0)
        self.assertEqual(ticks, model)
    
    
    def test_format(self):
        """Tests whether labels are formatted correctly."""
        
        # test small values
        ticker = pero.LinTicker(start=0, end=10)
        
        label = ticker.format(0.2)
        self.assertEqual(label, "0")
        
        label = ticker.format(2.7)
        self.assertEqual(label, "3")
        
        # test custom string format
        formatter = pero.StrFormatter(template="{:0.1f} u")
        ticker = pero.LinTicker(start=0, end=10, formatter=formatter)
        
        label = ticker.format(0.2)
        self.assertEqual(label, "0.2 u")
        
        # test custom function format
        formatter = pero.FuncFormatter(func=lambda x:"{:.3f} u".format(x))
        ticker = pero.LinTicker(start=0, end=10, formatter=formatter)
        
        label = ticker.format(0.2)
        self.assertEqual(label, "0.200 u")
        
        # test scientific notation
        ticker = pero.LinTicker(start=0, end=1e7)
        ticker.formatter.sci_notation = True
        
        label = ticker.format(2700000)
        self.assertEqual(label, "3e+06")
        
        # test suffix
        ticker = pero.LinTicker(start=0, end=1e7)
        ticker.formatter.sci_notation = True
        ticker.formatter.hide_suffix = True
        ticker.formatter.suffix_template = "10^{:0.0f}"
        
        suffix = ticker.suffix()
        self.assertEqual(suffix, "10^7")
    
    
    def test_beautify(self):
        """Tests whether beautify function works correctly."""
        
        ticker = pero.LinTicker()
        
        # test ascending
        beautified = ticker.beautify(0.201, 0.99)
        self.assertEqual(beautified, (0.2, 1))
        
        # test descending
        beautified = ticker.beautify(0.99, 0.201)
        self.assertEqual(beautified, (1, 0.2))


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import ui
from ... drawing import Graphics
from .. view import Control
from . view import UIView


class UIViewer(ui.View):
    """Simple graphics viewer application."""
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of UIViewer."""
        
        # init base
        super().__init__(**kwargs)
        self.name = "Pero"
        
        # init view
        self._view = None
    
    
    def set_size(self, size):
        """
        Sets app window size.
        
        Args:
            size: (float, float)
                App window size as (width, height).
        """
        
        self.width = size[0]
        self.height = size[1]
    
    
    def set_title(self, title):
        """
        Sets app window title.
        
        Args:
            title: str
                App window title.
        """
        
        self.name = title or ""
    
    
    def set_content(self, content):
        """
        Sets content to draw.
        
        Args:
            content: pero.UIView, pero.Control or pero.Graphics
                A content to set.
        """
        
        # clean sizer
        if self._view is not None:
            self.remove_subview(self._view)
        
        # init view
        if isinstance(content, UIView):
            self._view = content
        
        elif isinstance(content, Control):
            self._view = UIView()
            self._view.set_control(content)
        
        elif isinstance(content, Graphics):
            self._view = UIView()
            self._view.set_control(Control(graphics=content))
        
        else:
            message = "Unknown content type! -> %s" % type(content)
            raise TypeError(message)
        
        # add to sizer
        self.add_subview(self._view)
        
        # set layout
        self._view.flex = "WH"
    
    
    def layout(self):
        """Called when view is resized."""
        
        # set size to child
        if self._view is not None:
            self._view.frame = (0, 0, self.width, self.height)
    
    
    def refresh(self):
        """Redraws graphics."""
        
        if self._view is not None:
            self._view.refresh()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *


class Formatter(PropertySet):
    """
    This class mainly serves as a base class for all specific formatters. It can
    be also used alone just to convert given value into a string using the
    native implementation of provided value type (i.e. __str__ method).
    
    Each formatter has the 'domain' and 'precision' properties. They are used to
    prepare formatting according to actual data range and required precision
    (e.g. axis step size). If the formatter is used as part of an axis ticker,
    these properties are set automatically.
    
    Sometimes it might be handy to remove specific part from all the labels and
    put it for example at the end of the axis title (e.g. Axis Title [10^5]).
    Such string should then be available by overriding the 'suffix' method.
    
    Properties:
        
        domain: int, float, None or UNDEF
            Specifies the maximum absolute value of current range.
        
        precision: int, float, None or UNDEF
            Specifies the minimum required precision to be kept.
    """
    
    domain = NumProperty(UNDEF, dynamic=False, nullable=True)
    precision = NumProperty(UNDEF, dynamic=False, nullable=True)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats given value into a string using the native implementation of
        provided value type (i.e. __str__ method).
        
        This method should be overridden in derived classes to provide specific
        formatting mechanism.
        
        Args:
            value: any
                Value to be formatted.
        
        Returns:
            str
                Formatted label.
        """
        
        return str(value)
    
    
    def convert(self, value, *args, **kwargs):
        """
        Convert given value using current formatting. This is needed for certain
        types of formatters, which modify the value range (e.g. BytesFormatter).
        
        This method should be overridden in derived classes to provide specific
        conversion mechanism.
        
        Args:
            value: any
                Value to be scaled.
        
        Returns:
            any
                Converted value.
        """
        
        return value
    
    
    def invert(self, value, *args, **kwargs):
        """
        Inverts given value using current formatting. This is needed for certain
        types of formatters, which modify the value range (e.g. BytesFormatter).
        
        This method should be overridden in derived classes to provide specific
        conversion mechanism.
        
        Args:
            value: any
                Value to be inverted.
        
        Returns:
            any
                Inverted value.
        """
        
        return value
    
    
    def suffix(self, *args, **kwargs):
        """
        Gets current suffix (e.g. "10^5" or "kHz").
        
        This method should be overridden in derived classes to provide specific
        logic for creating the suffix.
        
        Returns:
            str
                Labels suffix.
        """
        
        return ""



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# import main objects
from . undefined import UNDEF
from . prop import Property
from . propset import PropertySet, Include, PROP_SPLITTER

# import additional properties
from . typed import FuncProperty, EnumProperty, RangeProperty, BoolProperty
from . typed import IntProperty, FloatProperty, NumProperty, StringProperty
from . typed import SequenceProperty, ListProperty, TupleProperty, SetProperty
from . typed import DictProperty, QuadProperty
from . special import ColorProperty, PaletteProperty, GradientProperty
from . special import DashProperty

# import property mixes
from . mixes import ColorProperties, AngleProperties
from . mixes import LineProperties, FillProperties, TextProperties



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for ordinal scale."""
    
    
    def test_scale(self):
        """Tests whether scale works correctly."""
        
        in_range = [1, 2, 3]
        out_range = ["red", "green", "blue"]
        
        scale = pero.OrdinalScale(
            in_range = in_range,
            out_range = out_range)
        
        self.assertEqual(scale.scale(1), "red")
        self.assertEqual(scale.scale(2), "green")
        self.assertEqual(scale.scale(3), "blue")
    
    
    def test_invert(self):
        """Tests whether invert works correctly."""
        
        in_range = [1, 2, 3]
        out_range = ["red", "green", "blue"]
        
        scale = pero.OrdinalScale(
            in_range = in_range,
            out_range = out_range)
        
        # check scale
        self.assertEqual(scale.invert("red"), 1)
        self.assertEqual(scale.invert("green"), 2)
        self.assertEqual(scale.invert("blue"), 3)
    
    
    def test_default(self):
        """Tests whether default value works correctly."""
        
        in_range = [1, 2, 3]
        out_range = ["red", "green", "blue"]
        default = "black"
        
        scale = pero.OrdinalScale(
            in_range = in_range,
            out_range = out_range,
            default = default)
        
        self.assertEqual(scale.scale(0), "black")
        self.assertEqual(scale.invert("black"), None)
    
    
    def test_implicit(self):
        """Tests whether implicit values work correctly."""
        
        out_range = ["red", "green", "blue"]
        default = None
        
        # disable implicit
        scale = pero.OrdinalScale(
            out_range = out_range,
            default = default,
            implicit = False)
        
        self.assertEqual(scale.scale(1), None)
        self.assertEqual(scale.invert("red"), None)
        
        # enable implicit
        scale = pero.OrdinalScale(
            out_range = out_range,
            default = default,
            implicit = True)
        
        self.assertEqual(scale.scale(1), "red")
        self.assertEqual(scale.scale(2), "green")
        self.assertEqual(scale.scale(1), "red")
        self.assertEqual(scale.scale(3), "blue")
        
        self.assertEqual(scale.invert("red"), 1)
        self.assertEqual(scale.invert("green"), 2)
        self.assertEqual(scale.invert("blue"), 3)
    
    
    def test_recycle(self):
        """Tests whether recycling works correctly."""
        
        in_range = [1, 2, 3, 4, 5, 6, 7]
        out_range = ["red", "green", "blue"]
        default = None
        
        # disable recycling
        scale = pero.OrdinalScale(
            in_range = in_range,
            out_range = out_range,
            default = default,
            recycle = False)
        
        self.assertEqual(scale.scale(1), "red")
        self.assertEqual(scale.scale(7), None)
        self.assertEqual(scale.invert("red"), 1)
        
        # enable recycling
        scale = pero.OrdinalScale(
            in_range = in_range,
            out_range = out_range,
            default = default,
            recycle = True)
        
        self.assertEqual(scale.scale(1), "red")
        self.assertEqual(scale.scale(4), "red")
        self.assertEqual(scale.scale(5), "green")
        self.assertEqual(scale.scale(6), "blue")
        self.assertEqual(scale.scale(7), "red")
        
        self.assertEqual(scale.invert("red"), 1)
        self.assertEqual(scale.invert("green"), 2)
        self.assertEqual(scale.invert("blue"), 3)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import cairo
from ... enums import *


CAIRO_LINE_CAP = {
    LINE_CAP_BUTT: cairo.LineCap.BUTT,
    LINE_CAP_SQUARE: cairo.LineCap.SQUARE,
    LINE_CAP_ROUND: cairo.LineCap.ROUND}

CAIRO_LINE_JOIN = {
    LINE_JOIN_BEVEL: cairo.LineJoin.BEVEL,
    LINE_JOIN_MITER: cairo.LineJoin.MITER,
    LINE_JOIN_ROUND: cairo.LineJoin.ROUND}

CAIRO_LINE_STYLE = {
    LINE_STYLE_CUSTOM: LINE_STYLE_CUSTOM,
    LINE_STYLE_SOLID: LINE_STYLE_SOLID,
    LINE_STYLE_DOTTED: DASH_VALUES_DOTTED,
    LINE_STYLE_DASHED: DASH_VALUES_DASHED,
    LINE_STYLE_DASHDOTTED: DASH_VALUES_DASHDOTTED}

CAIRO_FILL_RULE = {
    FILL_RULE_EVENODD: cairo.FillRule.EVEN_ODD,
    FILL_RULE_WINDING: cairo.FillRule.WINDING}

CAIRO_FONT_STYLE = {
    FONT_STYLE_NORMAL: cairo.FontSlant.NORMAL,
    FONT_STYLE_ITALIC: cairo.FontSlant.ITALIC}

CAIRO_FONT_WEIGHT = {
    FONT_WEIGHT_NORMAL: cairo.FontWeight.NORMAL,
    FONT_WEIGHT_LIGHT: cairo.FontWeight.NORMAL,
    FONT_WEIGHT_BOLD: cairo.FontWeight.BOLD,
    FONT_WEIGHT_BLACK: cairo.FontWeight.BOLD,
    FONT_WEIGHT_HEAVY: cairo.FontWeight.BOLD,
    FONT_WEIGHT_SEMIBOLD: cairo.FontWeight.BOLD,
    FONT_WEIGHT_MEDIUM: cairo.FontWeight.NORMAL,
    FONT_WEIGHT_ULTRALIGHT: cairo.FontWeight.NORMAL,
    FONT_WEIGHT_THIN: cairo.FontWeight.NORMAL}

CAIRO_FONT_FAMILY = {
    FONT_FAMILY_SERIF: FONT_FACE_SERIF,
    FONT_FAMILY_SANS: FONT_FACE_SANS,
    FONT_FAMILY_MONO: FONT_FACE_MONO}

CAIRO_RASTER_TYPES = {
    '.bmp': 'BMP',
    '.gif': 'GIF',
    '.jpg': 'JPEG',
    '.jpeg': 'JPEG',
    '.png': 'PNG',
    '.tif': 'TIFF',
    '.tiff': 'TIFF'}

CAIRO_VECTOR_TYPES = {
    '.eps': 'EPS',
    '.pdf': 'PDF',
    '.svg': 'SVG'}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. enums import *
from .. properties import *
from . scale import Scale


class LevelScale(Scale):
    """
    This type of scale tool is used to convert continuous input values into a
    predefined categorical threshold-based output. The input range should
    contain thresholds for each predefined level (maximum value allowed). The
    output range defines the final values representing particular level
    (e.g. color, name etc.).
    
    Properties:
        
        in_range: (float,)
            Specifies the maximum threshold for each level.
        
        out_range: (any,)
            Specifies the output levels values.
    """
    
    in_range = TupleProperty((), intypes=(int, float), dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of LevelScale."""
        
        super().__init__(**overrides)
        
        # init index
        self._levels_idx = None
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_levels_scale_property_changed)
    
    
    def scale(self, value, *args, **kwargs):
        """
        Returns corresponding output level for given input value.
        
        Args:
            value: float or (float,)
                Input value to be converted into level.
        
        Returns:
            any
                Corresponding level.
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.scale, value))
        
        # get index of the same or nearest bigger
        idx = 0
        hi = min(len(self.in_range), len(self.out_range))
        
        while idx < hi:
            mid = (idx + hi) // 2
            if self.in_range[mid] >= value:
                hi = mid
            else:
                idx = mid + 1
        
        # out of range
        if idx >= len(self.out_range):
            return None
        
        # get level
        return self.out_range[idx]
    
    
    def invert(self, value, *args, **kwargs):
        """
        Returns maximum threshold value for given level.
        
        Args:
            value: any or (any,)
                Level value.
        
        Returns:
            float
                Maximum threshold of the level.
        """
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.invert, value))
        
        # init index map
        if self._levels_idx is None:
            self._levels_idx = {}
            
            for i, item in enumerate(self.out_range):
                self._levels_idx[item] = i
        
        # get index
        idx = self._levels_idx.get(value, None)
        
        # check index
        if idx is None or idx >= len(self.in_range):
            return None
        
        # inside
        return self.in_range[idx]
    
    
    def _on_levels_scale_property_changed(self, evt=None):
        """Called after a property has changed."""
        
        # check in_range
        if evt is None or evt.name in 'in_range':
            if self.in_range is None or self.in_range is UNDEF:
                self.in_range = ()
        
        # check out_range
        if evt is None or evt.name == 'out_range':
            self._levels_idx = None
            if self.out_range is None or self.out_range is UNDEF:
                self.out_range = ()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import py5
from ... enums import *


PY5_LINE_CAP = {
    LINE_CAP_BUTT: py5.PROJECT,
    LINE_CAP_ROUND: py5.ROUND,
    LINE_CAP_SQUARE: py5.SQUARE}

PY5_LINE_JOIN = {
    LINE_JOIN_MITER: py5.MITER,
    LINE_JOIN_ROUND: py5.ROUND,
    LINE_JOIN_BEVEL: py5.BEVEL}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# define values
NONE = 'none'

DEG = 'deg'
RAD = 'rad'

LEFT = 'left'
RIGHT = 'right'
TOP = 'top'
BOTTOM = 'bottom'
CENTER = 'center'
MIDDLE = 'middle'
START = 'start'
END = 'end'

BEFORE = 'before'
AFTER = 'after'

EVENODD = 'evenodd'
WINDING = 'winding'

N = 'n'
NW = 'nw'
NE = 'ne'
S = 's'
SW = 'sw'
SE = 'se'
W = 'w'
E = 'e'
C = 'c'

HORIZONTAL = 'h'
VERTICAL = 'v'

INSIDE = 'inside'
OUTSIDE = 'outside'

NATURAL = 'natural'
FACEIN = 'facein'
FACEOUT = 'faceout'
FOLLOW = 'follow'

AUTO = 'auto'
CUSTOM = 'custom'

TRANS = 'trans'
SOLID = 'solid'
DOTTED = '.'
DASHED = '-'
DASHDOTTED = '-.'

BUTT = 'butt'
SQUARE = 'square'
BEVEL = 'bevel'
MITER = 'miter'
ROUND = 'round'

NORMAL = 'normal'
ITALIC = 'italic'
LIGHT = 'light'
BOLD = 'bold'
BLACK = 'black'
HEAVY = 'heavy'
SEMIBOLD = 'semibold'
MEDIUM = 'medium'
ULTRALIGHT = 'ultralight'
THIN = 'thin'

SERIF = 'serif'
SANS = 'sans'
MONO = 'mono'

JSON = 'json'
SVG = 'svg'
QT = 'qt'
WX = 'wx'
CAIRO = 'cairo'
MUPDF = 'mupdf'
PYTHONISTA = 'pythonista'
PY5 = 'py5'

DAYS = 'd'
HOURS = 'h'
MINUTES = 'm'
SECONDS = 's'
MSECONDS = 'ms'
USECONDS = 'us'
NSECONDS = 'ns'

CEIL = 'ceil'
FLOOR = 'floor'
HALFUP = 'halfup'
